"""
Main file of the project
"""

from omnetpypy import Experiment

if __name__ == '__main__':

    config_file = "config.yaml"
    topology_file = "topology.yaml"

    experiment = Experiment(config_file=config_file)
    experiment.run_simulations()



from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.quantum_node import QuantumNode
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.link_controller import LinkController

__all__ = ['QuantumNode', 'LinkController']


r"""
This module contains a set of global configuration variables for the simulation,
such as the engine to use, and a set of functions to make the configuration more user-friendly.
"""

import csv

import pandas as pd
import pkg_resources

from omnetpypy import utilities, sim_log, parser
from omnetpypy.backends.simpy_connector import SimPyConnector


class Simulation:
    r"""
    This class is used to run a single simulation using the specified configuration. It takes care of reading the
    running the simulation and returning the recorded metrics. Users should not instantiate this class directly,
    but use the :class:`~omnetpypy.simulation.Experiment` to manage multiple simulations.

    All parameters are also stored as attributes.

    Parameters
    ----------
    engine : str
        The simulation engine to use. Only "simpy" is implemented at the moment.
    seed_set : list of int
        The seed set for the repetition.
    repetition : int
        The repetition index.
    metrics : list of :class:`~omnetpypy.utilities.FutureMetric`
        The list of metrics to collect.
    yaml_directory : str
        The path of the directory that contains all the yaml files.
    until : int or float
        The time until which the simulation will run.
    log_level : str
        The log level for the simulation. See :func:`~omnetpypy.sim_log.log_to_console`.
    time_unit : str
        The time unit for the simulation. See :func:`~omnetpypy.utilities.time_unit_factor`.
    output_dir : str
        The output directory for the simulation.
    global_params : dict
        A dictionary of global parameters. These parameters are available to all the entities in the simulation.


    Attributes
    ----------
    rng : :class:`~omnetpypy.utilities.MultiRandom`
        The random number generator for this simulation, already initialized. It is made available to all the entities
        in the simulation through the :class:`~omnetpypy.backends.connector.Connector`.
    connector : :class:`~omnetpypy.backends.Connector`
        The connector to the simulation engine. Its subclass depends on the chosen engine.
    network : :class:`~omnetpypy.front_end.compound_module.CompoundModule`
        The network to simulate, parsed from the YAML file "network.yaml".

    Raises
    ------
    NotImplementedError
        If the passed ``engine`` is not implemented.
    """

    def __init__(self, engine, seed_set, repetition, metrics, yaml_directory, until, log_level, time_unit, output_dir,
                 global_params):
        self.engine = engine
        self.seed_set = seed_set
        self.repetition_idx = repetition
        self.until = until
        self.log_level = log_level
        self.time_unit = time_unit
        self.output_dir = output_dir

        self.global_params = global_params.copy()
        # dictionary of global parameters

        self.rng = utilities.MultiRandom(seeds=seed_set)
        self.connector = None

        # set log level
        sim_log.log_to_console(level=self.log_level)

        self.metrics = metrics
        # list of Metrics

        if engine == "simpy":
            self.connector = SimPyConnector(simulation=self, metrics=metrics, output_dir=output_dir,
                                            repetition=repetition)
        else:
            raise NotImplementedError("Engine not implemented")

        # parse yaml files
        self.network = parser.parse_yaml_directory(yaml_directory, self)

    def start(self):
        r"""
        start the simulation and collect the metrics.

        Returns
        -------
        dict
            A dictionary of dictionaries with the collected metrics.
        """
        self.connector.start_simulation(until=self.until)
        collected_data = {}
        # collect metrics
        for metric in self.metrics:
            if self.output_dir is not None:
                # dump the remaining samples in memory to the temporary file
                self.connector.dump_metric(metric.name)

                # compute the statistics we need
                filename = f"{self.output_dir}/.{metric.name}_vector_rep{self.repetition_idx}.csv"

                # check metric type
                if metric.type == "number":
                    collected_data[metric.name] = utilities.get_metrics_from_csv(metric, filename)
                else:
                    collected_data[metric.name] = {}

        return collected_data

    def time(self):
        """
        Return the current simulation time.

        Returns
        -------
        int or float
            The current simulation time.

        """
        return self.connector.get_time()

    @property
    def time_unit_factor(self):
        r"""
        The factor to convert the simulation time unit to seconds.

        See Also
        --------
        :func:`~omnetpypy.utilities.time_unit_factor`

        Returns
        -------
        int or float
            The factor to convert the time unit to seconds.

        """
        return utilities.time_unit_factor(self.time_unit)


class Experiment:
    r"""
    This class is used to run a set of independent simulations using the same YAML configuration files.
    An instance of this class takes care of reading the configuration file, parsing it, running the simulations, and
    dumping the recorded metrics in the output files.

    Parameters
    ----------
    config_file : str
        The path to the configuration file, absolute or relative to the working directory.
        The configuration file is a YAML file with a rigid structure (see ``config`` attribute).

    Attributes
    ----------
    seed_sets : list of lists of int
        The seed sets for each repetition. Each repetition has a seed set of length "rngs_per_rep"
        from the configuration file.
    config : dict
        The configuration dictionary, directly parsed from the YAML configuration file.
        The dictionary has the following keys:

            - engine (``str``), optional:
                The simulation engine to use. Only "simpy" is supported at the moment.
                Defaults to "simpy".
            - repetitions (``int``), optional:
                The number of independent repetitions to run. Defaults to 1.
            - metrics (list of :class:`~omnetpypy.utilities.FutureMetric`), optional:
                The list of metrics to collect. Defaults to an empty list.
            - yaml_directory (``str``), optional:
                The path to the directory containing the yaml files to set the simulation.
                Relative to the configuration file directory. Defaults to "./" .
            - simulate_until (``float`` or `None`), optional:
                The time until which the simulation will run. If `None`, simulation runs as long as there are events.
                Defaults to `None`.
            - time_unit (``str``), optional:
                The time unit for the simulation. Defaults to "us" (microseconds).
            - log_level (``int`` or ``str``), optional:
                The log level for the simulation. See :func:`~omnetpypy.sim_log.set_log_level`.
                Defaults to ``logging.INFO``
            - output_dir (``str`` or `None`), optional:
                The output directory for the simulation, relative to the configuration file directory.
                Defaults to `None`, in which case no data will be stored.
        
    simulations_params : list of tuples
        The list of simulation parameters for each repetition. Each element of the list is a tuple with the
        following parameters:

            - engine (``str``):
                The simulation engine to use.
            - seed_set (list of int):
                The seed set for the repetition.
            - repetition (``int``):
                The repetition index.
            - metrics (list of :class:`~omnetpypy.utilities.FutureMetric`):
                The list of metrics to collect.
            - yaml_directory (``str``):
                The path to the directory containing the yaml files to set the simulation.
            - until (``float`` or `None`):
                The time until which the simulation will run. If `None`, simulation runs as long as there are events.
            - log_level (``int`` or ``str``):
                The log level for the simulation. See :func:`~omnetpypy.sim_log.set_log_level`.
            - time_unit (``str``):
                The time unit for the simulation.
            - output_dir (``str`` or `None`):
                The output directory for the simulation.
            - global_params (dict):
                A dictionary of global parameters.
                These parameters are available to all the entities in the simulation.

    output_dir : str or None
        The output directory for the simulation. If `None`, no data is stored.


    Examples
    --------
    >>> exp = Experiment(config_file="config.yaml")
    >>> exp.run_simulations()

    """

    def __init__(self, config_file):

        self.seed_sets = []

        # every repetition has its own seed set of length rngs_per_rep
        # seeds are taken from the seeds.csv file
        # with a single "seeds" column. We open and read the file here
        # but not all the seeds are used, only the first rngs_per_rep * repetitions

        self.config = parser.parse_yaml_file(config_file)

        # derive the configuration file directory from its path by removing everything after the last "/"
        config_dir = config_file[:config_file.rfind("/") + 1]

        engine = self.config.get("engine", "simpy")
        repetitions = self.config.get("repetitions", 1)
        yaml_path = config_dir + self.config.get("yaml_directory", "./")
        rngs_per_rep = self.config.get("num_rngs", 1)
        until = self.config.get("simulate_until", None)
        log_level = self.config.get("log_level", "info")
        time_unit = self.config.get("time_unit", "us")

        global_params = self.config.get("global_params", {})

        # set log level
        sim_log.log_to_console(level=log_level)

        # so we iterate over the first rngs_per_rep * repetitions seeds
        # and we store them in a list of lists, where each list of seeds
        # is the seed set for a repetition
        seeds_path = pkg_resources.resource_filename('omnetpypy', 'data/seeds.csv')

        with open(seeds_path, "r") as f:
            reader = csv.reader(f)
            # use the reader sequentially on the first rngs_per_rep * repetitions rows
            for _ in range(repetitions):
                self.seed_sets.append([])
                for _ in range(rngs_per_rep):
                    self.seed_sets[-1].append(int(next(reader)[0]))

        metrics_raw = self.config.get("metrics", [])
        metrics = []

        for metric in metrics_raw:  # set up metric collection
            name = metric.get("name")
            metric_type = metric.get("type")
            collect = metric.get("collect", "number")

            all_collects = ["vector", "mean", "std", "min", "max", "median", "percentiles", "count", "var"]
            all_types = ["number", "str", "dict"]

            assert all([collect_type in all_collects for collect_type in collect]), f"collect must be one of {all_collects}"
            assert metric_type in all_types, f"type must be one of {all_types}"

            params_collect = {k: False for k in all_collects}

            true_params = {k: True for k in collect}
            params_collect.update(true_params)

            params = {"name": name,
                      "type": metric_type}
            params.update(params_collect)

            if metric_type != "number":
                assert len([k for k, v in params_collect.items() if v]) == 1 and params_collect["vector"], \
                    "only vector collect type must be set for nun-number typed metrics"

            columns = None
            if metric_type == "dict":
                columns = metric.get("columns")
                assert columns is not None, "columns must be provided for dict type metrics"
                assert isinstance(columns, list), "columns must be a list"
                assert all(isinstance(c, str) for c in columns), "columns must be a list of strings"
                assert len([k for k, v in params_collect.items() if v]) == 1 and params_collect["vector"], \
                    "only vector collect type must be True for dict type metrics"
                assert "timestamp" not in columns, "timestamp cannot be a column name, it is added automatically"

            params["columns"] = columns + ["timestamp"] if columns is not None else None

            metrics.append(utilities.FutureMetric(**params))

        if "output_dir" in self.config:
            output_dir = config_dir + self.config["output_dir"]
        else:
            output_dir = None
        self.output_dir = output_dir

        self.simulations_params = [(engine, self.seed_sets[i], i, metrics, yaml_path,
                                    until, log_level, time_unit, output_dir, global_params)
                                   for i in range(repetitions)]

    def run_simulations(self):
        r"""
        Run the simulations.
        """

        import multiprocessing

        max_processes = self.config.get("max_processes", 1)
        # check if max_processes is greater than the number of cpu cores
        if max_processes > 1:
            max_processes = min(multiprocessing.cpu_count(), max_processes)

        # if max_processes is 1, we just run the simulations sequentially
        if max_processes == 1:
            collected = []
            for sim_params in self.simulations_params:
                collected.append(_start_sim(sim_params))

        # run the simulations in parallel with a ProcessPoolExecutor
        else:
            with multiprocessing.Pool(max_processes) as pool:
                collected = pool.map(_start_sim, self.simulations_params)

        # every element of collected is a dictionary of dictionaries with the same keys.
        # we want to merge them into a single dictionary, where the values
        # are lists, except for the "name" key and the "vector" key, because the latter
        # is a pandas DataFrame and we will return a single merged DataFrame

        # first we create a dictionary
        merged = {}

        # now we iterate over the collected data
        for metric in collected[0].keys():
            merged[metric] = {}

        # now we iterate over the collected data
        for data in collected:
            for metric in data.keys():
                for key, value in data[metric].items():
                    if key != "vector":
                        if key not in merged[metric]:
                            merged[metric][key] = []
                        merged[metric][key].append(value)

        out_dir = self.output_dir
        # write the data to csv, two files for each metric: "metric_name.csv" and "metric_name_vector.csv"
        if out_dir is not None:
            for metric in merged.keys():
                # we write the statistics to a file if merged[metric] is not empty
                if merged[metric]:
                    df = pd.DataFrame(merged[metric])
                    df.to_csv(f"{out_dir}/{metric}.csv", index=False)

            for metric in merged.keys():
                # we open the temporary files and merge them into a single file. We add a "repetition" column
                max_reps = self.config.get("repetitions", 1)
                for i in range(max_reps):
                    if i == 0:
                        with open(f"{out_dir}/.{metric}_vector_rep{i}.csv", "r") as f:
                            # add the repetition column
                            df = pd.read_csv(f)
                            if max_reps > 1:
                                df["repetition"] = i
                            df.to_csv(f"{out_dir}/{metric}_vector.csv", index=False)

                    else:
                        with open(f"{out_dir}/.{metric}_vector_rep{i}.csv", "r") as f:
                            # add the repetition column
                            df = pd.read_csv(f)
                            if max_reps > 1:
                                df["repetition"] = i
                            df.to_csv(f"{out_dir}/{metric}_vector.csv", mode="a", header=False, index=False)
            # remove the temporary files
            import os
            for metric in merged.keys():
                for i in range(self.config.get("repetitions", 1)):
                    os.remove(f"{out_dir}/.{metric}_vector_rep{i}.csv")


def _start_sim(sim_params):
    # defined here to be picklable
    engine, seed_set, repetition, metrics, yaml_path, until, log_level, time_unit, output_dir, global_params = sim_params
    sim = Simulation(engine, seed_set, repetition, metrics, yaml_path, until, log_level, time_unit, output_dir, global_params)
    return sim.start()



from omnetpypy import SimpleModule, Message, sim_log
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir import messages
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.messages import RoutablePacket, EntanglementGenPacket, \
    FlowsInformationPacket, FlowDeletionPacket
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.utility import sanitize_flow_descriptors


class LinkController(SimpleModule):
    """
    This class models a Link Controller, placed in the middle of a link.
    """

    def __init__(self, name, identifier, t_clock):
        port_names = ["lc0", "lc1"]

        super().__init__(name, identifier, port_names)
        self.t_clock = t_clock

        self._flow_probabilities = {}
        # dictionary of flow probabilities, one for each flow, indexed by flow_id. They must sum to 1.

        self._flow_attempt_probabilities = {}
        # dictionary of flow attempt probabilities, one for each flow, indexed by flow_id. They must NOT sum to 1.

        self._flow_directions = {}

        self._trigger_msg = Message(["trigger attempt"], header="trigger")

        self._cur_lle_id = 0

    def initialize(self, step=0):
        if step == 0:
            """
            Schedule the first attempt to entangle the flows
            """
            first_attempt_time = self.sim_context.rng.random(generator=0) * self.t_clock
            self.schedule_message(self._trigger_msg, delay=first_attempt_time)

            # access global parameters and get the flows information
            flow_descriptors = self.sim_context.global_params["flow_descriptors"]
            flow_descriptors = sanitize_flow_descriptors(flow_descriptors)
            # dict of flow descriptors, indexed by flow_id, each one containing the following fields (for now):
            # - flow_id
            # - source
            # - destination
            # - path (list of node names, including source and destination)
            # - success_probs (list of success probabilities for each link in the path)
            # - direction (upstream or downstream)

            message = messages.FlowsInformationPacket(destination=self.name, flows=flow_descriptors)
            self._handle_flows_information(message)

    def _attempt_entanglement(self):

        # In this first implementation, we update the flow probabilities right before each attempt, because we have
        # perfect, real-time information on the queues at the adjacent nodes. In future implementations, we will use a
        # more realistic approach where we update the flow probabilities at regular intervals with reduced information.
        self._update_flow_probabilities()

        queues_info = self._get_queues_info()

        # check whether the queues are empty. if so, we don't attempt entanglement
        if queues_info[0].is_empty() and queues_info[1].is_empty():
            self.schedule_message(self._trigger_msg, delay=self.t_clock)
            return

        rng = self.sim_context.rng

        """# first we pick a flow to attempt entanglement for using the flow probabilities

        # now we get the flows for which queues are not empty
        flows = [flow_id for flow_id in self._flow_probabilities.keys() if not queues_info[0].is_empty(flow_id)]

        # we filter out flows whose direction is not upstream
        flows = [flow_id for flow_id in flows if self._flow_directions[flow_id] == "upstream"]

        flows_down = [flow_id for flow_id in self._flow_probabilities.keys() if not queues_info[1].is_empty(flow_id)]

        # we filter out flows whose direction is not downstream
        flows_down = [flow_id for flow_id in flows_down if self._flow_directions[flow_id] == "downstream"]

        flows += flows_down
        # remove duplicates
        assert len(list(set(flows))) == len(flows)

        # check whether flows is empty. if so, we don't attempt entanglement
        if len(flows) == 0:
            self.schedule_message(self._trigger_msg, delay=self.t_clock)
            return

        # derive the probabilities for these flows by normalizing the flow probabilities
        flows_probabilities = [self._flow_probabilities[flow_id] for flow_id in flows]
        summed = sum(flows_probabilities)
        flows_probabilities = [p / summed for p in flows_probabilities]

        # we pick the flow_id using the flow probabilities and the default rng (generator = 0)
        flow_id = rng.choices(sequence=flows,
                              weights=flows_probabilities,
                              k=1,
                              generator=0)[0]"""

        # flip a coin to decide which queue to peek
        left_is_owner = False

        # get queues length
        q0_len = queues_info[0].length()
        q1_len = queues_info[1].length()

        left_prob = q0_len / (q0_len + q1_len)

        if rng.random(generator=0) < left_prob:
            queue = 0
            req, oldest_time = queues_info[0].peek_request(out_port="q1", policy="OLDEST")
            left_is_owner = True
        else:
            queue = 1
            req, oldest_time = queues_info[1].peek_request(out_port="q0", policy="OLDEST")

        if req is None:
            # try the other queue
            queue = 1 - queue
            req, oldest_time = queues_info[queue].peek_request(out_port="q0" if queue == 1 else "q1", policy="OLDEST")
            left_is_owner = not left_is_owner

        if req is None:
            self.schedule_message(self._trigger_msg, delay=self.t_clock)
            return
        flow_id = req.flow_id

        # use a geometric distribution to determine the number of attempts needed
        attempts = rng.geometric(p=self._flow_attempt_probabilities[flow_id], generator=0)

        # create a self message
        self_msg = Message([flow_id, left_is_owner], header="entanglement ready")
        # wait for the number of attempts
        self.schedule_message(self_msg, delay=(attempts-1) * self.t_clock)

    def _handle_successful_entanglement(self, message):

        flow_id = message.fields[0]
        left_is_owner = message.fields[1]

        lle_id = self.name + "-" + str(self._cur_lle_id)
        # debug: check that the request we picked is indeed the oldest in the queue for that port name
        # print(f"Here {self.name} for {lle_id} : Request {req} is the oldest ({oldest_time}) in the queue on port q{1 - queue}")
        # print(queues_info[queue]._requests)
        self.send(EntanglementGenPacket(flow_id=flow_id, lle_id=lle_id, sender_name=self.name,
                                        owner=left_is_owner), "lc0")
        self.send(EntanglementGenPacket(flow_id=flow_id, lle_id=lle_id, sender_name=self.name,
                                        owner=not left_is_owner), "lc1")
        self._cur_lle_id += 1

        # we schedule the next attempt

        self.schedule_message(self._trigger_msg, delay=self.t_clock)

    def _get_queues_info(self):
        """
        Get the information on request queues stored on adjacent nodes available to this link controller.
        """

        # The first implementation is a magic trick where we crawl the network to find references to the adjacent
        # nodes and we directly access their queues. In future implementations, we will use a more realistic approach
        # where we send messages to the adjacent nodes asking for their queue status.
        # TODO

        # veeeeeery ugly and unsafe
        node_right = self.ports["lc1"].connected_port.parent.ports["B"].connected_port.parent
        node_left = self.ports["lc0"].connected_port.parent.ports["A"].connected_port.parent

        return node_left.req_queue, node_right.req_queue

    def _update_flow_probabilities(self):
        """
        Update the flow probabilities based on the information on the queues at the adjacent nodes.
        """
        # First implementation: use _get_queues_info to get the queues status and update the flow probabilities based
        # on the number of requests in the queues.
        # TODO
        pass

    def handle_message(self, message, port_name):

        if isinstance(message, FlowsInformationPacket) and message.destination == self.name:
            sim_log.debug(f"{self.name} received flows information with {len(message.flows)} flows.",
                          time=self.sim_context.time())
            self._handle_flows_information(message)
            return

        if isinstance(message, FlowDeletionPacket) and message.destination == self.name:
            sim_log.debug(f"{self.name} received flow deletion for flow {message.flow_id}.",
                          time=self.sim_context.time())
            self._handle_flow_deletion(message)
            return

        if port_name is not None and isinstance(message, RoutablePacket):
            # if this is not the destination, forward the packet on the other port
            if message.destination != self.name:
                # print(f"{self.name} forwarding message {message} to {port_name} at time {self.sim_context.time()}")
                self.send(message, "lc1" if port_name == "lc0" else "lc0")
                return

        elif port_name is None:  # self message, we need to attempt entanglement
            if "header" in message.meta and message.meta["header"] == self._trigger_msg.meta["header"]:
                self._attempt_entanglement()
            elif "header" in message.meta or message.meta["header"] == "entanglement ready":
                self._handle_successful_entanglement(message)
            else:
                raise ValueError("Unknown self message received")

    def _handle_flows_information(self, message):
        for flow in message.flows:
            if self.name not in flow["path"]:
                continue
            flow_id = flow["flow_id"]
            self._flow_probabilities[flow_id] = 1 / len(message.flows)  # begin with equal probabilities for all flows

            # store the flow direction
            self._flow_directions[flow_id] = flow["direction"]

            # find our position on the flow path
            path = flow["path"]
            idx = path.index(self.name)
            # the path is node0---(link_first_half)---link_controller---(link_second_half)---node1...
            # the link controller is attached to link_first_half and link_second_half
            # find link position (first + second halves are one link) on the path that this link controller controls
            link_pos = int((idx - 1) / 2)
            # get the success probability for the link
            success_prob = flow["success_probs"][link_pos]
            # update the flow probabilities
            self._flow_attempt_probabilities[flow_id] = success_prob


    def _handle_flow_deletion(self, message):
        flow_id = message.flow_id
        if flow_id in self._flow_probabilities:
            del self._flow_probabilities[flow_id]
            del self._flow_attempt_probabilities[flow_id]
            del self._flow_directions[flow_id]
        else:
            sim_log.warning(f"Flow {flow_id} not found in the flow probabilities of {self.name}.",
                            time=self.sim_context.time())
            raise ValueError(f"Flow {flow_id} not found in the flow probabilities of {self.name}.")
        return






from quantum_bell_api.decoherence import depolarize_rate
from quantum_bell_api.swapping import swap
from quantum_bell_api.utility import epr_pair, get_werner_state
from omnetpypy import SimpleModule, sim_log, Message
import omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.messages as messages
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.aqm_controller import PIController
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.congestion_controller import WindowCongestionController, \
    RateCongestionController
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.queues import RequestQueue, LLEManager
from omnetpypy.examples.qnum_congestion_ctrl_aqm_bidir.utility import sanitize_flow_descriptors


class QuantumNode(SimpleModule):
    """
    This class models a Quantum Repeater along a repeater chain

    Parameters
    ----------
    name : str
        The name of the module
    identifier : int
        The identifier of the module
    storage_qbits_per_port : int or None, optional
        The number of  storage qubits available on each port. If None, the module will assume an infinite storage
        capacity. Default is None.
    """

    def __init__(self, name, identifier, storage_qbits_per_port=None, decoherence_rate=0.):

        port_names = ["q0", "q1"]
        # first dilemma: do we use separate channels for classical packets  (c0, c1 ports) ?
        # rg0 is used to attach the RequestGenerator

        super().__init__(name, identifier, port_names)
        self.flows_info = None
        self.lle_manager = LLEManager(port_names=["q0", "q1"])
        self.req_queue = RequestQueue()
        self.storage_qbits_per_port = storage_qbits_per_port
        self.decoherence_rate = decoherence_rate

        ################################
        # CONGESTION CONTROL VARIABLES #
        ################################
        self.congestion_controller = RateCongestionController()

        self.cur_req_ids = {}  # dictionary of current request ids, one for each flow, indexed by flow_id, used to
        # keep track of which id assign to the next generated request for each flow
        self.timeout_trigger_msg = Message(["timeout control trigger"], header="timeout control trigger")
        self.timeout_trigger_period = 20000  # us
        self.last_update_time = 0
        ################################
        # END OF CONGESTION CONTROL    #
        ################################

        ################################
        # ACTIVE QUEUE MGMT VARIABLES  #
        ################################
        self.aqm_controllers = None
        self.AQM_T_sampling_times = None
        self.AQM_update_msgs = None
        ################################
        # END OF AQM VARIABLES         #
        ################################

        ################################
        # NEW FLOW GENERATION          #
        ################################
        self.new_flow_trigger_msg = Message(["new flow trigger"], header="new flow trigger")
        self.new_flow_trigger_period = 8000000  # us
        self.delete_phase = False
        ################################
        # END OF NEW FLOW GENERATION    #
        ################################

        ################################
        # NEW REQUEST GENERATION       #
        ################################
        self.new_request_trigger_msgs = {}
        self.increase_request_rate_trigger_msgs = {}
        ################################
        # END OF NEW REQUEST GENERATION #
        ################################

    def initialize(self, step=0):
        if step == 0:
            # usually you can retrieve parameters by looking at self.parent attributes (parent is the compound module
            # containing this module)

            # access global parameters and get the flows information
            flow_descriptors = self.sim_context.global_params["flow_descriptors"]
            flow_descriptors = sanitize_flow_descriptors(flow_descriptors)
            # dict of flow descriptors, indexed by flow_id, each one containing the following fields (for now):
            # - flow_id
            # - source
            # - destination
            # - path (list of node names, including source and destination)
            # - success_probs (list of success probabilities for each link in the path)
            # - direction (upstream or downstream)

            message = messages.FlowsInformationPacket(destination=self.name, flows=flow_descriptors)
            self._handle_flows_information(message)

            self.schedule_message(self.timeout_trigger_msg, delay=self.timeout_trigger_period)

            self.aqm_controllers = {
                "upstream": PIController(),
                "downstream": PIController()
            }
            aqm_params = self.sim_context.global_params["aqm_params"]
            self.AQM_T_sampling_times = {
                "upstream": (self.aqm_controllers["upstream"].set_parameters(**aqm_params))*self.sim_context.time_unit_factor,
                "downstream": (self.aqm_controllers["downstream"].set_parameters(**aqm_params))*self.sim_context.time_unit_factor
            }
            self.AQM_update_msgs = {
                "upstream": Message(["downstream"], header="update AQM downstream"),
                "downstream": Message(["upstream"], header="update AQM upstream")
            }

            for direction in ["upstream", "downstream"]:
                self._schedule_aqm_update(direction)

            self.schedule_message(self.new_flow_trigger_msg, delay=self.new_flow_trigger_period)

    def _schedule_aqm_update(self, direction):
        """
        Schedule the next AQM update for the given direction
        """
        self.schedule_message(self.AQM_update_msgs[direction], delay=self.AQM_T_sampling_times[direction])

    def _get_success_prob(self, request, direction):
        """
        Get the success probability for the request to be forwarded on the given port
        """
        # get this node's position in the path
        path = self.flows_info[request.flow_id]["path"]
        # Remove elements at even indices
        path_no_link_controllers = [path[i] for i in range(len(path)) if i % 2 == 0]
        idx = path_no_link_controllers.index(self.name)
        if direction == "downstream":
            # we are going downstream
            return self.flows_info[request.flow_id]["success_probs"][idx - 1]
        elif direction == "upstream":
            # we are going upstream
            return self.flows_info[request.flow_id]["success_probs"][idx]
        else:
            raise ValueError(f"Unknown direction: {direction}")


    def handle_message(self, message, port_name):

        if isinstance(message, messages.FlowsInformationPacket) and message.destination == self.name:
            self._handle_flows_information(message)
            return

        if isinstance(message, messages.FlowDeletionPacket) and message.destination == self.name:
            self._handle_flow_deletion(message)
            return

        if "header" in message.meta and message.meta["header"] == "initialize requests":
            # we generate the first requests for each flow for which we are the source or the destination
            for flow_id in self.flows_info:
                if self.name == self.flows_info[flow_id]["source"] or self.name == self.flows_info[flow_id]["destination"]:
                    if not isinstance(self.congestion_controller, RateCongestionController):
                        self.generate_request(flow_id)
                    else:
                        # schedule the next request generation
                        self.schedule_message(self.new_request_trigger_msgs[flow_id],
                                              delay=self.congestion_controller.get_inter_request_gap(flow_id=flow_id))
                        # schedule the next increase in the flow rate
                        self.schedule_message(self.increase_request_rate_trigger_msgs[flow_id],
                                              delay=self.congestion_controller.estimated_rtt[flow_id])
            return

        if "header" in message.meta and message.meta["header"] == "new flow trigger":
            """self._generate_new_flow()
            self.schedule_message(self.new_flow_trigger_msg, delay=self.new_flow_trigger_period)"""
            return

        if "header" in message.meta and message.meta["header"] == "new requests trigger":

            if message.fields[0] not in self.flows_info:
                # the flow has been deleted
                del self.new_request_trigger_msgs[message.fields[0]]
                return

            self.generate_request(flow_id=message.fields[0])
            delay = self.congestion_controller.get_inter_request_gap(flow_id=message.fields[0])
            self.schedule_message(message, delay=delay)
            return

        if "header" in message.meta and message.meta["header"] == "increase flow rate trigger":
            if not isinstance(self.congestion_controller, RateCongestionController):
                raise ValueError("The global rate increase period is set but the congestion controller"
                                 "is not a RateCongestionController")

            if message.fields[0] not in self.flows_info:
                # the flow has been deleted
                del self.increase_request_rate_trigger_msgs[message.fields[0]]
                return

            flow_id = message.fields[0]
            self.congestion_controller.increase_congestion_knob(flow_id=message.fields[0],
                                                                current_time=self.sim_context.time())
            # get estimated rtt
            rtt = self.congestion_controller.estimated_rtt[flow_id]
            # schedule the next increase for the estimated rtt

            self.schedule_message(message, delay=rtt)
            return

        if port_name is None and message.meta["header"].startswith("update AQM"):
            direction = message.fields[0]
            out_port = "q0" if direction == "downstream" else "q1"
            current_q_len = self.req_queue.weighted_length(out_port=out_port)
            self.aqm_controllers[direction].update(q=current_q_len)
            self._schedule_aqm_update(direction)
            return

        if isinstance(message, messages.RoutablePacket) and message.destination != self.name:
            # if this is not the destination, forward the packet on the other port
            self.send(message, "q1" if port_name == "q0" else "q0")
            return

        if isinstance(message, messages.EntanglementRequestPacket):
            self._handle_entanglement_request(message, port_name)
            return

        if isinstance(message, messages.EntanglementGenPacket):
            self._handle_new_lle(message, port_name)
            return

        if isinstance(message, messages.EntanglementGenAcknowledgement):
            self._handle_req_ack(message, port_name)
            return

        if "header" in message.meta and message.meta["header"] == "timeout control trigger":
            self.collect_timeouts()
            return

        else:
            raise ValueError(f"Unknown message received: {message} at node {self.name}")

    def _handle_flows_information(self, message):
        """
        This method is called when a FlowsInformationPacket is received
        """
        # among all flows, we only keep those that are relevant to this node
        # (i.e. the flows where the current node is part of the path)
        relevant_flows = {}
        for flow in message.flows:
            if self.name in flow["path"]:
                relevant_flows[flow["flow_id"]] = flow

        flow_info = {}
        for flow_id in relevant_flows:
            flow = relevant_flows[flow_id]
            next_port = {"downstream": "q0", "upstream": "q1"}
            if self.name == flow["source"]:
                self.congestion_controller.setup_congestion_control(flow, current_time=self.sim_context.time())
                self.cur_req_ids[flow_id] = 0
                self.new_request_trigger_msgs[flow_id] = Message([flow_id], header="new requests trigger")
                self.increase_request_rate_trigger_msgs[flow_id] = Message([flow_id], header="increase flow rate trigger")
            if self.name == flow["destination"]:
                self.congestion_controller.setup_congestion_control(flow, current_time=self.sim_context.time(),
                                                                    is_source=False)
                self.cur_req_ids[flow_id] = 1000000
                self.new_request_trigger_msgs[flow_id] = Message([flow_id], header="new requests trigger")
                self.increase_request_rate_trigger_msgs[flow_id] = Message([flow_id], header="increase flow rate trigger")


            if self.name != flow["destination"]:
                next_hop_up = flow["path"][flow["path"].index(self.name) + 2]
            else:
                next_hop_up = None
            if self.name != flow["source"]:
                next_hop_down = flow["path"][flow["path"].index(self.name) - 2]
            else:
                next_hop_down = None
            # next hop in the path, +2 because the path includes the link controllers

            flow_info[flow_id] = {
                "next_port": next_port,
                "source": flow["source"],
                "destination": flow["destination"],
                "next_hop_down": next_hop_down,
                "next_hop_up": next_hop_up,
                "success_probs": flow["success_probs"],
                "path": flow["path"]
            }

        was_init = False
        if self.flows_info is None:
            self.flows_info = flow_info
            was_init = True
        else:
            self.flows_info.update(flow_info)

        sim_log.debug(f"{self.name} received flows information with {len(relevant_flows)} relevant flows.",
                      time=self.sim_context.time())

        # we generate the first requests for each flow for which we are the source or the destination
        # but we wait for a little time to let the other nodes initialize
        if was_init:
            self.schedule_message(Message(["initialize requests"], header="initialize requests"), delay=10)
        else:
            for flow_id in flow_info:
                if self.name == flow_info[flow_id]["source"] or self.name == flow_info[flow_id]["destination"]:
                    if not isinstance(self.congestion_controller, RateCongestionController):
                        self.generate_request(flow_id)
                    else:
                        # schedule the next request generation
                        self.schedule_message(self.new_request_trigger_msgs[flow_id],
                                              delay=self.congestion_controller.get_inter_request_gap(flow_id=flow_id))
                        # schedule the next increase in the flow rate
                        self.schedule_message(self.increase_request_rate_trigger_msgs[flow_id],
                                              delay=self.congestion_controller.estimated_rtt[flow_id])

    def _handle_flow_deletion(self, message):
        """
        This method is called when a FlowDeletionPacket is received
        """
        flow_id = message.flow_id
        if flow_id in self.flows_info:

            # if we are the source of the flow, we have to delete the flow from the congestion controller
            # and cancel the scheduled messages
            if self.name == self.flows_info[flow_id]["source"] or self.name == self.flows_info[flow_id]["destination"]:
                self.congestion_controller.delete_flow(flow_id)

            # delete all requests in the queue for the flow and all the LLEs
            self.req_queue.delete_requests(flow_id)
            self.lle_manager.delete_lles(flow_id)

            # delete the flow information
            del self.flows_info[flow_id]
            sim_log.debug(f"Flow {flow_id} deleted at node {self.name}", time=self.sim_context.time())
        else:
            sim_log.error(f"Flow {flow_id} not found at node {self.name}", time=self.sim_context.time())
            raise ValueError(f"Flow {flow_id} not found at node {self.name}")


    def generate_request(self, flow_id):
        """
        Generate a request for the given flow_id
        """
        # generate a new request
        req_id = self.cur_req_ids[flow_id]
        self.cur_req_ids[flow_id] += 1

        # determine the direction and the next hop destination depending on whether we are the source or the destination
        if self.name == self.flows_info[flow_id]["source"]:
            direction = "upstream"
            destination = self.flows_info[flow_id]["next_hop_up"]
        else:
            direction = "downstream"
            destination = self.flows_info[flow_id]["next_hop_down"]

        request_pkt = messages.EntanglementRequestPacket(destination=destination, flow_id=flow_id, req_id=req_id,
                                                         lle_id=None, gen_time=self.sim_context.time())

        # set the success probabilities as a meta field
        request_pkt.meta["success_probs"] = self.flows_info[flow_id]["success_probs"][:]
        request_pkt.meta["direction"] = direction

        # send the request to ourself
        self._handle_new_request(request_pkt)

        # let the congestion controller know that a new request has been generated
        self.congestion_controller.handle_new_request_in_flight(req_id=req_id, flow_id=flow_id,
                                                                current_time=self.sim_context.time())


    def collect_timeouts(self):
        """
        Collect the timeouts for the requests in flight and update the congestion windows for the involved flows
        """
        self.congestion_controller.collect_timeouts(self.sim_context.time())
        self.schedule_message(self.timeout_trigger_msg, delay=self.timeout_trigger_period)

        if self.last_update_time + 20000 <= self.sim_context.time():
            self.last_update_time = self.sim_context.time()
            # log some simulation updates for the user
            sim_log.debug(f"Node {self.name} has currently {len(self.req_queue)} requests in queue and {len(self.lle_manager)} LLEs",
                          time=self.sim_context.time())
            # for all flows of which we are the source, we log the current generation probability
            for flow_id in self.flows_info:
                if self.name == self.flows_info[flow_id]["source"] and isinstance(self.congestion_controller, RateCongestionController):
                    p_gen = self.congestion_controller.get_inter_request_gap(flow_id=flow_id)
                    sim_log.debug(f"Node {self.name} has an IRG = {p_gen} us for flow {flow_id}",
                                  time=self.sim_context.time())
                if self.name == self.flows_info[flow_id]["destination"] and isinstance(self.congestion_controller, RateCongestionController):
                    p_gen = self.congestion_controller.get_inter_request_gap(flow_id=flow_id)
                    sim_log.debug(f"Node {self.name} has an IRG = {p_gen} us for flow {flow_id}",
                                  time=self.sim_context.time())

    def _generate_new_flow(self, num_flows=3):
        """
        Generate a new flow by copying the first one of which this node is the source
        """
        # if total number of flows is equal or greater than 13, we delete the last num_flows flows for which we are the source
        if len(self.flows_info) >= 13:
            self.delete_phase = True

        if len(self.flows_info) == 4 and self.delete_phase:
            self.delete_phase = False

        if self.delete_phase:
            flow_ids = list(self.flows_info.keys())
            # sort the flow_ids in descending order
            flow_ids.sort(reverse=True)
            to_delete = num_flows
            for f_id in flow_ids:
                if self.name == self.flows_info[f_id]["source"] and to_delete > 0:
                    # delete flow message
                    flow = self.flows_info[f_id]
                    for destination in flow["path"]:
                        del_flow_info = messages.FlowDeletionPacket(destination=destination, flow_id=f_id)
                        if destination == self.name:
                            self._handle_flow_deletion(del_flow_info)
                            continue
                        if flow["direction"] == "upstream":
                            self.send(del_flow_info, port_name="q1")
                        else:
                            self.send(del_flow_info, port_name="q0")
                    to_delete -= 1
            return


        # we just copy the first flow for which we are the source
        flow_id = None
        for f_id in self.flows_info:
            if self.name == self.flows_info[f_id]["source"]:
                flow_id = f_id
                break

        if flow_id is None:
            return

        flow = self.flows_info[flow_id]

        for _ in range(num_flows):
            new_flow = flow.copy()
            if new_flow["direction"] == "upstream":
                new_flow["flow_id"] = len(self.flows_info)
            else:
                new_flow["flow_id"] = len(self.flows_info) + num_flows
            self.flows_info[new_flow["flow_id"]] = new_flow

            # send the new flow information to the other nodes
            for destination in new_flow["path"]:
                new_flow_info = messages.FlowsInformationPacket(destination=destination, flows=[new_flow])

                if destination == self.name:
                    self._handle_flows_information(new_flow_info)
                    continue

                # it will be routed to the destination node
                if new_flow["direction"] == "upstream":
                    self.send(new_flow_info, port_name="q1")
                else:
                    self.send(new_flow_info, port_name="q0")

            sim_log.debug(f"Node {self.name} generated a new flow {new_flow['flow_id']}",
                          time=self.sim_context.time())

    def _handle_entanglement_request(self, message, port_name):
        """
        This method is called when an EntanglementRequestPacket is received
        """

        flow_id = message.flow_id

        assert flow_id in self.flows_info, f"Flow {flow_id} not found in the flows information"
        # assert port_name != self.flows_info[flow_id]["next_port"], "Received a request from the wrong port"

        """
        sim_log.debug(f"Request {message.req_id} received for flow {flow_id}. Here is node {self.name}",
                      time=self.sim_context.time())
        """

        # first of all we pop the first success probability from the request meta because it won't be used anymore
        message.meta["success_probs"].pop(0)

        # first thing: check whether the request's LLE is still available, it might have been dropped due to storage
        # qubit shortage
        lle, lle_time = self.lle_manager.peek_from_req(request=message, raise_error=False)
        if lle is None:
            # the lle has been dropped because of storage qubit shortage. We have to drop the request
            return

        # emit the queue size metric
        if self.name == "qn2":
            self.emit_metric("queue_size", self.req_queue.weighted_length(out_port="downstream"))

        elif self.name == "qn3":
            self.emit_metric("queue_size_free", self.req_queue.weighted_length(out_port="downstream"))

        direction = message.meta["direction"]

        # now let's check whether we should mark the request as congested
        if not message.is_congested():
            aqm = self.aqm_controllers[direction]
            marking_prob = aqm.get_marking_probability()
            random_number = self.sim_context.rng.random(generator=flow_id)
            # use the i-th random number generator offered by the sim_context
            # to check whether the request should be marked as congested
            if random_number < marking_prob:
                message.mark_congested()


        if ((direction == "upstream" and self.name == self.flows_info[flow_id]["destination"]) or
                (direction == "downstream" and self.name == self.flows_info[flow_id]["source"])):
            # we are the destination, we are done :)
            # pop the lle the request refers to from the available lles and emit 1
            lle, lle_time = self.lle_manager.pop_from_req(request=message, raise_error=True)

            """sim_log.warning(f"Request {message.req_id} satisfied for flow {flow_id} with fidelity {fidelity}. Here is node {self.name}",
                          time=self.sim_context.time())"""

            # get the wait time of the lle
            wait_time = self.sim_context.time() - lle_time
            # decohere the state
            self._decohere_state(message, wait_time, get_werner_state(fidelity=1.))

            # generate and send the acknowledgement
            destination = self.flows_info[flow_id]["source"] if direction == "upstream" else self.flows_info[flow_id]["destination"]
            ack = messages.EntanglementGenAcknowledgement(req_id=message.req_id, flow_id=flow_id,
                                                          destination=destination,
                                                          congested=message.is_congested(),
                                                          gen_time=message.gen_time,
                                                          qstate=message.meta["qstate"],
                                                          ack_time=self.sim_context.time())
            self.send(ack, port_name=port_name)

            """
            sim_log.debug(f"Request {message.req_id} satisfied for flow {flow_id}. Here is node {self.name}",
                          time=self.sim_context.time())
            """

            self.emit_metric("rendezvous_node", int(self.name[2:]))

            return


        # Now, we check whether there is another request in the queue for the same flow in the opposite direction
        # if so, we swap the two requests
        other_direction = "upstream" if direction == "downstream" else "downstream"
        other_port = "q0" if port_name == "q1" else "q1"
        other_request, _ = self.req_queue.pop_request(flow_id=flow_id, out_port=port_name, policy=RequestQueue.OLDEST)
        if other_request is not None:
            # there is a request in the queue for the same flow in the opposite direction
            # we swap the two requests

            # we pop the lle for the other request
            other_lle, other_lle_time = self.lle_manager.pop_from_req(request=other_request, raise_error=True)
            # we pop the lle for the current request
            lle, lle_time = self.lle_manager.pop_from_req(request=message, raise_error=True)

            wait_time = self.sim_context.time() - lle_time
            other_wait_time = self.sim_context.time() - other_lle_time

            # decohere both states
            self._decohere_state(other_request, other_wait_time, get_werner_state(fidelity=1.))
            self._decohere_state(message, wait_time, get_werner_state(fidelity=1.))

            # swap the two requests
            self._decohere_state(other_request, wait_time=0., other_pair=message.meta["qstate"])
            message.meta["qstate"] = other_request.meta["qstate"]
            # now both states have been updated and match

            # generate the acknowledgments for both requests
            destination_a = self.flows_info[flow_id]["source"] if direction == "upstream" else self.flows_info[flow_id]["destination"]
            destination_b = self.flows_info[flow_id]["source"] if other_direction == "upstream" else self.flows_info[flow_id]["destination"]

            # determine which destination is further away from this node on the path
            path = self.flows_info[flow_id]["path"]
            idx_a = path.index(destination_a)
            idx_b = path.index(destination_b)
            this_idx = path.index(self.name)
            if abs(idx_a - this_idx) < abs(idx_b - this_idx):  # to determine who collects stats
                destination_a, destination_b = destination_b, destination_a
                other_request, message = message, other_request
                other_direction, direction = direction, other_direction
                other_port, port_name = port_name, other_port



            ack_a = messages.EntanglementGenAcknowledgement(req_id=message.req_id, flow_id=flow_id,
                                                            destination=destination_a,
                                                            congested=message.is_congested(),
                                                            gen_time=message.gen_time,
                                                            qstate=message.meta["qstate"],
                                                            ack_time=self.sim_context.time())
            ack_b = messages.EntanglementGenAcknowledgement(req_id=other_request.req_id, flow_id=flow_id,
                                                            destination=destination_b,
                                                            congested=other_request.is_congested(),
                                                            gen_time=other_request.gen_time,
                                                            skip_stats=True)

            # send the acknowledgments
            self.send(ack_a, port_name=port_name)
            self.send(ack_b, port_name=other_port)

            # log a bunch of debug info about who is sending what to whom
            """sim_log.debug(
                f"Request {message.req_id} swapped with request {other_request.req_id} for flow {flow_id}. Here is node {self.name}."
                f" Request {message.req_id} sent to {destination_a} through port {port_name} and request {other_request.req_id} sent to {destination_b} through port {other_port}",
                time=self.sim_context.time())"""

            self.emit_metric("rendezvous_node", int(self.name[2:]))

            return

        # we have to check whether we have a lle to swap with the request
        # if so, we swap, update the message information and forward the request to the next node
        next_port = self.flows_info[flow_id]["next_port"][direction]
        next_hop = self.flows_info[flow_id]["next_hop_down"] if direction == "downstream" else self.flows_info[flow_id]["next_hop_up"]
        if self.lle_manager.is_empty(port_name=next_port, flow_id=flow_id):
            # just append the request to the corresponding queue
            success_prob = self._get_success_prob(message, direction=message.meta["direction"])
            self.req_queue.add_request(message, next_port, self.sim_context.time())
            return

        # there is at least a lle for this flow that can be swapped with the request
        # pop the lle for the request
        # pop the youngest suitable lle
        other_lle, other_lle_time = self.lle_manager.pop_lle(port_name=next_port, flow_id=flow_id, owner=True,
                                                             policy=LLEManager.YOUNGEST)

        if other_lle is None:
            # we don't own any lle for this flow, we just append the request to the corresponding queue
            success_prob = self._get_success_prob(message, direction=message.meta["direction"])
            self.req_queue.add_request(message, next_port, self.sim_context.time())
            return

        lle, lle_time = self.lle_manager.pop_from_req(request=message, raise_error=True)

        # update request
        wait_time = self.sim_context.time() - other_lle_time
        message.update_request(lle_id=other_lle.lle_id, wait_time=wait_time,
                               destination=next_hop)

        # decohere the quantum state described within the request
        self._decohere_state(message, wait_time)

        # send the message
        self.send(message, port_name=next_port)

        # emit queueing time for the request (only intermediate repeater)
        if self.name == "qn1":
            self.emit_metric("queuing_time", 0.0)

    def _decohere_state(self, message, wait_time, other_pair=None):
        # decohere the quantum state described within the request
        wait_time_seconds = wait_time / self.sim_context.time_unit_factor
        rate = message.meta["src_decoherence_rate"] + self.decoherence_rate
        new_state = depolarize_rate(message.meta["qstate"], rate, wait_time_seconds)
        if other_pair is None:
            other_pair = get_werner_state(fidelity=1.)
        new_state = swap(new_state, other_pair, eta=1., p_2=1.)
        message.meta["qstate"] = new_state


    def _handle_new_request(self, message):
        flow_id = message.flow_id

        # if we are not the source we ignore the request
        if (flow_id not in self.flows_info or
                (self.name != self.flows_info[flow_id]["source"] and self.name != self.flows_info[flow_id]["destination"])):
            raise ValueError(f"New request for which we are not the source: {message} at node {self.name}")


        # initialize the epr pair state that will be tracked and updated after every swap
        epr_pair_initial = get_werner_state(fidelity=1.)
        message.meta["qstate"] = epr_pair_initial
        message.meta["src_decoherence_rate"] = self.decoherence_rate

        """
        sim_log.debug(f"Request generated for flow {flow_id}. Here is node {self.name} with a queue empty? {self.req_queue.is_empty()}",
                      time=self.sim_context.time())
        """

        # now let's check whether we should mark the request as congested
        if not message.is_congested():
            direction = message.meta["direction"]
            aqm = self.aqm_controllers[direction]
            marking_prob = aqm.get_marking_probability()
            random_number = self.sim_context.rng.random(generator=flow_id)
            # use the i-th random number generator offered by the sim_context
            # to check whether the request should be marked as congested
            if random_number < marking_prob:
                message.mark_congested()


        # we are the source, we have to check whether we have a lle to associate with the request
        # if so, we associate the lle and send the request to the next node
        direction = message.meta["direction"]
        next_port = self.flows_info[flow_id]["next_port"][direction]
        next_hop = self.flows_info[flow_id]["next_hop_down"] if direction == "downstream" else self.flows_info[flow_id][
            "next_hop_up"]
        if self.lle_manager.is_empty(port_name=next_port, flow_id=flow_id):
            # just append the request to the corresponding queue

            # check that the number of requests in queue in this direction is less than 100
            if self.req_queue.weighted_length(out_port=next_port) >= 2*self.storage_qbits_per_port:
                # drop the request
                return

            self.req_queue.add_request(message, out_port=next_port, time=self.sim_context.time())
            return

        # there is at least an lle for this flow that can be assigned to the request
        # pop the lle
        lle, lle_time = self.lle_manager.pop_lle(port_name=next_port, flow_id=message.flow_id, owner=True,
                                                 policy=LLEManager.YOUNGEST)

        if lle is None:
            # we don't own any lle for this flow, we just append the request to the corresponding queue
            self.req_queue.add_request(message, out_port=next_port, time=self.sim_context.time())
            return

        # update request
        message.update_request(lle_id=lle.lle_id, wait_time=None,
                               destination=next_hop)

        # send the message
        self.send(message, port_name=next_port)


    def _handle_new_lle(self, message, port_name):
        """
        This method is called when an EntanglementGenPacket is received
        """
        flow_id = message.flow_id
        # we have to check whether there is a request to be swapped with this lle
        # if so, we swap, update the message information and forward the request to the next node

        # first of all we have to check whether the flow is still active
        if flow_id not in self.flows_info:
            # the flow has been deleted
            # do nothing
            return

        # first we check if we are the owner of the lle
        if not message.meta["owner"]:
            # we are not the owner, we just have to store the lle
            self.add_lle(message, port_name)
            return

        if self.req_queue.length(flow_id=flow_id, out_port=port_name) == 0:  # no requests for this flow :(
            # append
            self.add_lle(message, port_name)
            """# if flow is upstream and port_name is q1 we raise an error
            if self.flows_info[flow_id]["direction"] == "upstream" and port_name == "q1":
                print("Maybe About to raise error. The interested queue is", self.req_queue._requests)
                print("The LLE msg is ", message)
                raise ValueError(f"{self.name} Received a new lle for flow {flow_id} on {port_name} for which there are no requests")
            # if flow is downstream and port_name is q0 we raise an error
            if self.flows_info[flow_id]["direction"] == "downstream" and port_name == "q0":
                print("About to raise error. The interested queue is", self.req_queue._requests)
                print("The LLE msg is ", message)
                raise ValueError(f"{self.name} Received a new lle for flow {flow_id} on {port_name} for which there are no requests")"""
            return

        # now we are sure that the new lle is in the flow direction we were waiting for and there is a request for it

        # pop the request from the queue
        request, request_time = self.req_queue.pop_request(flow_id, out_port=port_name, policy=RequestQueue.OLDEST)

        # there is at least a request for this flow on this port
        direction = request.meta["direction"]
        next_port = port_name
        next_hop = self.flows_info[flow_id]["next_hop_down"] if direction == "downstream" else self.flows_info[flow_id][
            "next_hop_up"]

        # Now we have to check whether we are the source of the flow
        if ((direction == "upstream" and self.name == self.flows_info[flow_id]["source"]) or
                (direction == "downstream" and self.name == self.flows_info[flow_id]["destination"])):
            # in this case we just have to associate the new lle to the request and send it to the next node
            request.update_request(lle_id=message.lle_id, wait_time=None,
                                   destination=next_hop)

            self.send(request, port_name=next_port)
            return

        # we are not the source, we have to pop the lle associated with the request
        # pop lle associated to the popped request
        other_lle, other_lle_time = self.lle_manager.pop_from_req(request=request, raise_error=True)

        # swap the lles
        # update the request message by adding the swapped lle wait time
        wait_time = self.sim_context.time() - other_lle_time

        # update the request message
        request.update_request(lle_id=message.lle_id, wait_time=wait_time,
                               destination=next_hop)

        # decohere the quantum state described within the request
        self._decohere_state(request, wait_time)

        # send the message
        self.send(request, port_name=next_port)

        # emit queueing time for the request (only intermediate repeater)
        if self.name == "qn1":
            self.emit_metric("queuing_time", self.sim_context.time() - request_time)

    def add_lle(self, lle, port_name):
        """
        Add an LLE to the manager. If there are no free storage qubits, we try to pop the oldest lle for the same flow
        and replace it with the new one. If there are no lles for the same flow, we pop the oldest lle
        for that port and replace it with the new one.

        Parameters
        ----------
        lle : EntanglementGenPacket
            The LLE to add
        port_name : str
            The name of the port on which the LLE is available
        """

        if self.storage_qbits_per_port is not None:
            if self.lle_manager.length(port_name) >= self.storage_qbits_per_port:
                # we have to pop the oldest lle for the same flow
                # if there are no lles for the same flow, we pop the oldest lle for that port
                # if there are no lles for that port, we raise an error
                lle_to_pop, _ = self.lle_manager.pop_lle(port_name=port_name, flow_id=lle.flow_id,
                                                      policy=LLEManager.OLDEST)
                if lle_to_pop is None:
                    # we have to pop the oldest lle for that port
                    lle_to_pop, _ = self.lle_manager.pop_lle(port_name=port_name, policy=LLEManager.OLDEST)
                    if lle_to_pop is None:
                        raise ValueError("No LLEs available for popping")

                # now we have to check whether the popped lle is associated with a request
                # if so, we have to drop the request
                request, _ = self.req_queue.pop_from_lle(lle_to_pop, raise_error=False)
                if request is not None:
                    # the request has been dropped
                    sim_log.warning(f"Request {request.req_id} dropped at node {self.name} due to storage qubit shortage",
                                    time=self.sim_context.time())
                self.lle_manager.add_lle(lle, port_name, self.sim_context.time())
                return

        self.lle_manager.add_lle(lle, port_name, self.sim_context.time())

    def _handle_req_ack(self, message, port_name):
        """
        This method is called when an EntanglementGenAcknowledgement is received
        """
        flow_id = message.flow_id

        # log the request acknowledgement
        """sim_log.debug(f"Request {message.req_id} acknowledged for flow {flow_id}. Here is node {self.name}",
                      time=self.sim_context.time())"""

        if "skip_stats" not in message.meta or not message.meta["skip_stats"]:
            ack_transmission_time = self.sim_context.time() - message.meta["ack_time"]
            qstate = message.meta["qstate"]
            # decohere the quantum state described within the request
            rate = 2*self.decoherence_rate  # 2 because both end nodes will decohere the state
            ack_transmission_time_seconds = ack_transmission_time / self.sim_context.time_unit_factor
            qstate = depolarize_rate(qstate, rate, ack_transmission_time_seconds)
            fidelity = qstate.a

            # emit the fidelity metric
            self.emit_metric("fidelity", fidelity)

            self.emit_metric(name="throughput", value=1)

            self.emit_metric("latency", self.sim_context.time() - message.gen_time)

        if flow_id not in self.flows_info:
            sim_log.warning(f"Received an acknowledgement for a flow not in the flows information: {message} at node {self.name}",
                            time=self.sim_context.time())
            return

        # first we check whether we are the source of the flow, otherwise we throw an error
        if self.name != self.flows_info[flow_id]["source"] and self.name != self.flows_info[flow_id]["destination"]:
            raise ValueError(f"Received an acknowledgement for a flow for which we are not"
                             f"the source or destination: {message} at node {self.name}")

        num_new_requests = self.congestion_controller.handle_ack(req_id=message.req_id, flow_id=flow_id,
                                                                 mark_congested=message.congested,
                                                                 current_time=self.sim_context.time(),
                                                                 time_sent=message.gen_time)

        # emit the congestion window for flow 0
        if flow_id == 0 and isinstance(self.congestion_controller, WindowCongestionController):
            self.emit_metric("congestion_window", self.congestion_controller.get_congestion_window(flow_id))

        elif flow_id == 0 and isinstance(self.congestion_controller, RateCongestionController):
            self.emit_metric("IRG", self.congestion_controller.get_inter_request_gap(flow_id))

        # we generate new requests for the flow
        for _ in range(num_new_requests):
            self.generate_request(flow_id)



from omnetpypy import Message


class RoutablePacket(Message):
    """
    This class models a classical packet that can be routed through a network.

    The only constraint to the packet is that it must have a specified destination in its meta.
    """

    def __init__(self, destination, fields, **meta):
        meta["destination"] = destination
        super().__init__(fields, **meta)

    @property
    def destination(self):
        return self.meta["destination"]

    @destination.setter
    def destination(self, value):
        self.meta["destination"] = value


class FlowsInformationPacket(RoutablePacket):
    r"""
    This class models a message that carries information about the flows in the network.

    Parameters
    ----------

    flows : list
        Flow descriptors, each one containing the following fields:
        <ul>
            <li>flow_id</li>
            <li>source</li>
            <li>destination</li>
            <li>path (list of node names, including link controllers)</li>
            <li>success_probs (list of success probabilities for each link in the path)</li>
            <li>request_rate (average, Poisson process)</li>
            <li>direction (upstream or downstream)</li>
        </ul>
    """

    HEADER = "FLOWS INFORMATION"

    def __init__(self, flows, destination, **meta):
        meta["header"] = FlowsInformationPacket.HEADER
        fields = [flows]
        super().__init__(destination, fields, **meta)

    @property
    def flows(self):
        return self.fields[0]

    def __copy__(self):
        new_meta = self.meta.copy()
        flows_cpy = [desc.copy() for desc in self.flows]
        return FlowsInformationPacket(flows=flows_cpy, **new_meta)


class FlowDeletionPacket(RoutablePacket):
    """
    This class models a message that carries information about a flow deletion in the network.

    Parameters
    ----------

    flow_id : int
        The identifier of the flow to be deleted
    """

    HEADER = "FLOW DELETION"

    def __init__(self, flow_id, destination, **meta):
        meta["header"] = FlowDeletionPacket.HEADER
        fields = [flow_id]
        super().__init__(destination, fields, **meta)

    @property
    def flow_id(self):
        return self.fields[0]

class EntanglementRequestPacket(RoutablePacket):

    HEADER = "ENTANGLEMENT REQUEST"

    def __init__(self, req_id, destination, flow_id, lle_id, gen_time, wait_times=None, **meta):
        if wait_times is None:
            wait_times = []
        super().__init__(destination, fields=[req_id, flow_id, lle_id, gen_time, wait_times],
                         header=EntanglementRequestPacket.HEADER, congested_ECN=False, **meta)

    @property
    def req_id(self):
        return self.fields[0]

    @property
    def flow_id(self):
        return self.fields[1]

    @property
    def lle_id(self):
        return self.fields[2]

    @lle_id.setter
    def lle_id(self, value):
        self.fields[2] = value

    @property
    def gen_time(self):
        return self.fields[3]

    @property
    def wait_times(self):
        return self.fields[4]

    @property
    def direction(self):
        if "direction" not in self.meta:
            return None
        return self.meta["direction"]

    def is_congested(self):
        """
        Whether the request has been marked with an Explicit Congestion Notification (ECN) or not.

        Returns
        -------
        bool
            True if the request has been marked as congested, False otherwise.
        """
        return self.meta["congested_ECN"]

    def mark_congested(self):
        """
        Mark the request as congested, using an Explicit Congestion Notification (ECN).
        """
        self.meta["congested_ECN"] = True

    def append_wait_time(self, wait_time):
        self.wait_times.append(wait_time)

    def __copy__(self):
        new_meta = self.meta.copy()
        del new_meta["destination"]
        del new_meta["header"]
        del new_meta["congested_ECN"]
        fields_cpy = self.fields[:]
        ret = EntanglementRequestPacket(req_id=fields_cpy[0], destination=self.destination,
                                         flow_id=fields_cpy[1], lle_id=fields_cpy[2], wait_times=fields_cpy[3][:],
                                         **new_meta)
        if self.is_congested():
            ret.mark_congested()

        return ret

    def update_request(self, lle_id=None, wait_time=None, destination=None):
        """
        Update the request with the new lle_id, wait_time and destination. Usually done right before forwarding the
        request to the next node.

        Parameters
        ----------
        lle_id : int or None, optional
            The new lle_id
        wait_time : float or None, optional
            The wait time for the last swapped lle, appended to the wait_times list
        destination : str or None, optional
            The new destination for the request (next hop)
        """
        if lle_id is not None:
            self.lle_id = lle_id
        if wait_time is not None:
            self.append_wait_time(wait_time)
        if destination is not None:
            self.destination = destination


class EntanglementGenPacket(Message):
    """
    This class models the message sent by link controllers to nodes, encapsulating the entanglement generation
    for a given flow
    """

    HEADER = "ENTANGLEMENT GENERATION"

    def __init__(self, flow_id, lle_id, sender_name, **meta):
        meta["header"] = EntanglementGenPacket.HEADER
        fields = [flow_id, lle_id, sender_name]
        super().__init__(fields, **meta)

    @property
    def flow_id(self):
        return self.fields[0]

    @property
    def lle_id(self):
        return self.fields[1]

    @property
    def sender_name(self):
        return self.fields[2]


class EntanglementGenAcknowledgement(RoutablePacket):
    """
    This class models the message sent by nodes to link controllers, acknowledging the entanglement generation
    for a given flow
    """

    HEADER = "ENTANGLEMENT GEN ACK"

    def __init__(self, destination, flow_id, req_id, gen_time, congested=False, **meta):
        fields = [flow_id, req_id, gen_time, congested]
        super().__init__(fields=fields, destination=destination, header=EntanglementGenAcknowledgement.HEADER,
                         **meta)

    @property
    def flow_id(self):
        return self.fields[0]

    @property
    def req_id(self):
        return self.fields[1]

    @property
    def gen_time(self):
        return self.fields[2]

    @property
    def congested(self):
        return self.fields[3]

    @congested.setter
    def congested(self, value):
        self.fields[3] = value




"""
In this module we implement the RequestQueue class, which is a simple wrapper around a list of requests. It is used to
store the requests that are waiting to be processed by QuantumNodes.
"""


class RequestQueue:
    """
    This class is a simple wrapper around a list of requests. It is used to store the requests that are waiting to be
    processed by QuantumNodes.
    """

    # create an enumerator for the popping policy (oldest or youngest)
    OLDEST = 0
    YOUNGEST = 1

    def __init__(self):
        self._requests = {}

    def add_request(self, request, out_port, time):
        """
        Add a request to the queue

        Parameters
        ----------
        request : EntanglementRequestPacket
            The request to add
        out_port : str
            The name of the port where the request is to be forwarded
        time : float
            The simulation time the request was added to the queue
        """
        if out_port not in self._requests:
            self._requests[out_port] = []

        self._requests[out_port].append((request, time))

    def _merged_requests(self):
        # merge all queues into a single one ordered by time
        queue = []

        """for q in self._requests.values():  # simple merge but not computationally efficient (O(nlogn))
            queue.extend(q)
        queue = sorted(queue, key=lambda x: x[1])"""

        # implement a sorted merge (O(n))
        for port, q in self._requests.items():
            if len(queue) == 0:
                queue = [(elem[0], elem[1], port) for elem in q]
            else:
                i = 0
                j = 0
                merged = []
                while i < len(queue) and j < len(q):
                    if queue[i][1] < q[j][1]:
                        merged.append((queue[i][0], queue[i][1], queue[i][2]))
                        i += 1
                    else:
                        merged.append((q[j][0], q[j][1], port))
                        j += 1
                if i < len(queue):
                    rest = [(elem[0], elem[1], elem[2]) for elem in queue[i:]]
                    merged.extend(rest)
                if j < len(q):
                    rest = [(elem[0], elem[1], port) for elem in q[j:]]
                    merged.extend(rest)
                queue = merged
        return queue

    def pop_request(self, flow_id, out_port=None, policy=OLDEST):
        """
        Remove and return the first request in the queue

        Parameters
        ----------
        flow_id : int
            The flow id of the request to pop
        out_port : str or None, optional
            The name of the port to pop the request from. If None, any port will be considered. Default is None.
        policy : int or None, optional
            The policy to use when popping the request. It can be either OLDEST (0) or YOUNGEST (1). Default is OLDEST.

        Returns
        -------
        tuple
            A tuple with the request and the time it was added to the queue, or (None, None) if no request for that flow
            id was found
        """
        queue = None
        call_again = False
        if out_port is not None:
            if out_port not in self._requests:
                return None, None
            queue = self._requests[out_port]
        else:
            queue = self._merged_requests()
            call_again = True

        ret = None
        if policy == self.OLDEST:
            # pop the oldest request
            for i, entry in enumerate(queue):
                if entry[0].flow_id == flow_id:
                    ret = queue.pop(i)
                    break
        else:
            # pop the youngest request
            for i, entry in enumerate(queue[::-1]):
                if entry[0].flow_id == flow_id:
                    ret = queue.pop(-i - 1)
                    break

        if call_again and ret is not None:
            return self.pop_request(flow_id, out_port=ret[2], policy=policy)
        elif ret is not None:
            return ret
        return None, None

    def pop_from_lle(self, lle, out_port=None, raise_error=False):
        """
        Remove and return the request in the queue associated with the LLE

        Parameters
        ----------
        lle : EntanglementGenPacket
            The LLE to satisfy
        out_port : str or None, optional
            The name of the port to pop the request from. If None, any port will be considered. Default is None.
        raise_error : bool, optional
            Whether to raise an error if no request is found. Default is False.

        Returns
        -------
        tuple
            A tuple with the request and the time it was added to the queue, or (None, None) if no request was found
            and raise_error is False
        """
        call_again = False
        if out_port is not None:
            queue = self._requests[out_port]
        else:
            queue = self._merged_requests()
            call_again = True
        target = None
        for i, entry in enumerate(queue):
            if entry[0].lle_id == lle.lle_id:
                target = queue.pop(i)
                break

        if call_again and target is not None:
            return self.pop_from_lle(target[0], out_port=target[2], raise_error=raise_error)
        elif target is not None:
            return target
        elif raise_error:
            raise ValueError("No request found for the given LLE")
        return None, None

    def peek_request(self, flow_id=None, out_port=None, policy=OLDEST):
        """
        Return the first request in the queue without removing it

        Parameters
        ----------
        flow_id : int
            The flow id of the request to peek. If None, any flow id will be considered. Default is None.
        out_port : str
            The name of the port to peek the request from. If None, any port will be considered. Default is None.
        policy : int
            The policy to use when peeking the request. It can be either OLDEST (0) or YOUNGEST (1)

        Returns
        -------
        tuple
            A tuple with the request and the time it was added to the queue, or (None, None) if no request for that flow
            id was found
        """
        queue = None
        call_again = False
        if out_port is not None:
            if out_port not in self._requests:
                return None, None
            queue = self._requests[out_port]
        else:
            queue = self._merged_requests()
            call_again = True

        ret = None
        if policy == self.OLDEST:
            # pop the oldest request
            for i, entry in enumerate(queue):
                if flow_id is None or entry[0].flow_id == flow_id:
                    ret = queue[i]
                    break
        else:
            # pop the youngest request
            for i, entry in enumerate(queue[::-1]):
                if flow_id is None or entry[0].flow_id == flow_id:
                    ret = queue[-i - 1]
                    break

        if ret is not None:
            return ret[0], ret[1]

        return None, None

    def delete_requests(self, flow_id):
        """
        Delete all requests for a given flow id

        Parameters
        ----------
        flow_id : int
            The flow id of the requests to delete
        """

        for port_name in self._requests.keys():
            self._requests[port_name] = [req for req in self._requests[port_name] if req[0].flow_id != flow_id]

    def is_empty(self, flow_id=None):
        """
        Check whether the queue is empty for a given (optional) flow identifier

        Parameters
        ----------
        flow_id : int or None, optional
            The flow id to check for requests. If None, the method will check if the queue is empty. Default is None.

        Returns
        -------
        bool
            True if there are no requests for a given flow, False otherwise. If no flow id is given, return True if
            there are no requests in the queue, False otherwise.
        """
        if flow_id is not None:
            for queue in self._requests.values():
                if len(queue) > 0:
                    for req, time in queue:
                        if req.flow_id == flow_id:
                            return False
            return True
        else:
            for queue in self._requests.values():
                if len(queue) > 0:
                    return False
            return True

    def __len__(self):
        tot_len = 0
        for queue in self._requests.values():
            tot_len += len(queue)
        return tot_len

    def length(self, flow_id=None, out_port=None):
        """
        Return the total number of requests in the queue for a given flow id and/or port name

        Parameters
        ----------
        flow_id : int or None, optional
            The flow id of the requests to count. If None, all requests will be counted. Default is None.
        out_port : str or None, optional
            The name of the port to count the requests from. If None, all ports will be considered. Default is None.

        Returns
        -------
        int
            The total number of requests in the queue for the given flow id
        """

        if out_port is not None:
            if out_port not in self._requests:
                return 0

            if flow_id is not None:
                return len([req for req in self._requests[out_port] if req[0].flow_id == flow_id])

            return len(self._requests[out_port])
        else:
            if flow_id is not None:
                return len([req for queue in self._requests.values() for req in queue if req[0].flow_id == flow_id])

            return len(self)

    def weighted_length(self, out_port):
        """
        Get the total number of requests belonging to any flow with the specified direction.
        The weight of each request is 1/success_prob.

        Parameters
        ----------
        out_port : string
            The output port to consider

        Returns
        -------
        float
            The weighted queue length for the given output port
        """
        tot_len = 0.
        tot_len_unweighted = 0
        if out_port not in self._requests:
            return 0.
        for req, time in self._requests[out_port]:
            success_prob = 1 if "success_prob" not in req.meta else req.meta["success_prob"]
            tot_len += 1. / success_prob
            tot_len_unweighted += 1
        return tot_len_unweighted

class LLEManager:
    """
    This class is a wrapper around a list of LLEs. It is used to store the LLEs that are available to be used by
    QuantumNodes.
    """

    # create an enumerator for the popping policy (oldest or youngest)
    OLDEST = 0
    YOUNGEST = 1

    def __init__(self, port_names):
        self._lles = {port_name: [] for port_name in port_names}

    def add_lle(self, lle, port_name, time):
        """
        Add an LLE to the manager

        Parameters
        ----------
        lle : EntanglementGenPacket
            The LLE to add
        port_name : str
            The name of the port on which the LLE is available
        time : float
            The simulation time the LLE was added to the manager
        """
        self._lles[port_name].append((lle, time))

    def pop_from_req(self, request, raise_error=True):
        """
        Remove and return the LLE in the manager associated with the request

        Parameters
        ----------
        request : EntanglementRequestPacket
            The request to satisfy
        raise_error : bool, optional
            Whether to raise an error if no LLE is found. Default is True.

        Returns
        -------
        tuple
            A tuple with the LLE and the time it was added to the manager, or (None, None) if no LLE was found
        """
        for port_name, lles in self._lles.items():
            for i, (lle, time) in enumerate(lles):
                if lle.flow_id == request.flow_id and lle.lle_id == request.lle_id:
                    return self._lles[port_name].pop(i)
        if raise_error:
            raise ValueError("No LLE found for the given request")
        return None, None

    def peek_from_req(self, request, raise_error=True):
        """
        Return the LLE in the manager associated with the request

        Parameters
        ----------
        request : EntanglementRequestPacket
            The request to satisfy
        raise_error : bool, optional
            Whether to raise an error if no LLE is found. Default is True.

        Returns
        -------
        tuple
            A tuple with the LLE and the time it was added to the manager, or (None, None) if no LLE was found
        """
        for port_name, lles in self._lles.items():
            for lle, time in lles:
                if lle.flow_id == request.flow_id and lle.lle_id == request.lle_id:
                    return lle, time
        if raise_error:
            raise ValueError("No LLE found for the given request")
        return None, None

    def pop_lle(self, port_name, flow_id=None, owner=False, policy=OLDEST):
        r"""
        Pop an LLE from the manager available on the given port, according to the given policy

        Parameters
        ----------
        port_name : str
            The name of the port where the LLE is available
        flow_id : int or None, optional
            The flow id of the LLE to pop. If None, the method will pop the first LLE available on the port. Default is
            None.
        owner : bool, optional
            Whether the LLE should be owned by the QuantumNode that pops it. Default is False. If False, the LLE can either
            be owned by the QuantumNode or not.
        policy : int
            The policy to use when popping the LLE. It can be either OLDEST (0) or YOUNGEST (1)

        Returns
        -------
        tuple
            A tuple with the LLE and the time it was added to the manager, or (None, None) if no LLE was found

        """

        if policy == self.OLDEST:
            for i, (lle, time) in enumerate(self._lles[port_name]):
                if flow_id is None or lle.flow_id == flow_id:
                    if (not owner) or (owner and lle.meta["owner"]):
                        return self._lles[port_name].pop(i)
        elif policy == self.YOUNGEST:
            for i, (lle, time) in enumerate(self._lles[port_name][::-1]):
                if flow_id is None or lle.flow_id == flow_id:
                    if (not owner) or (owner and lle.meta["owner"]):
                        return self._lles[port_name].pop(-i - 1)
        else:
            raise ValueError("Invalid policy")

        return None, None

    def peek_lle(self, port_name, flow_id=None, owner=False, policy=OLDEST):
        """
        Peek at an LLE from the manager available on the given port, according to the given policy

        Parameters
        ----------
        port_name : str
            The name of the port where the LLE is available
        flow_id : int or None, optional
            The flow id of the LLE to pop. If None, the method will pop the first LLE available on the port. Default is
            None.
        owner : bool, optional
            Whether the LLE should be owned by the QuantumNode that pops it. Default is False. If False, the LLE can either
            be owned by the QuantumNode or not.
        policy : int
            The policy to use when popping the LLE. It can be either OLDEST (0) or YOUNGEST (1)

        Returns
        -------
        tuple
            A tuple with the LLE and the time it was added to the manager, or (None, None) if no LLE was found

        """

        if policy == self.OLDEST:
            for lle, time in self._lles[port_name]:
                if flow_id is None or lle.flow_id == flow_id:
                    if (not owner) or (owner and lle.meta["owner"]):
                        return lle, time
        elif policy == self.YOUNGEST:
            for lle, time in self._lles[port_name][::-1]:
                if flow_id is None or lle.flow_id == flow_id:
                    if (not owner) or (owner and lle.meta["owner"]):
                        return lle, time
        else:
            raise ValueError("Invalid policy")

        return None, None

    def delete_lles(self, flow_id):
        """
        Delete all LLEs for a given flow id

        Parameters
        ----------
        flow_id : int
            The flow id of the LLEs to delete
        """

        for port_name in self._lles.keys():
            self._lles[port_name] = [lle for lle in self._lles[port_name] if lle[0].flow_id != flow_id]


    def is_empty(self, port_name, flow_id=None):
        """
        Check whether the manager is empty for a given (optional) flow identifier

        Parameters
        ----------
        port_name : str
            The name of the port to check for LLEs
        flow_id : int or None, optional
            The flow id to check for LLEs. If None, the method will check if the manager is empty. Default is None.

        Returns
        -------
        bool
            True if there are no LLEs for a given flow, False otherwise. If no flow id is given, return True if
            there are no LLEs in the manager, False otherwise.
        """
        if flow_id is not None:
            return len([lle for lle in self._lles[port_name] if lle[0].flow_id == flow_id]) == 0
        else:
            return len(self._lles[port_name]) == 0

    def __len__(self):
        """
        Return the total number of LLEs in the manager

        Returns
        -------
        int
            The total number of LLEs in the manager
        """
        tot_len = 0
        for lles in self._lles.values():
            tot_len += len(lles)
        return tot_len

    def length(self, port_name):
        """
        Return the total number of LLEs in the manager for a given port

        Returns
        -------
        int
            The total number of LLEs in the manager for the given port
        """
        return len(self._lles[port_name])



"""
Main file for the ping pong simulation example.
"""

from omnetpypy import Experiment

if __name__ == '__main__':

    config_file = "ping_pong_config.yaml"

    experiment = Experiment(config_file=config_file)
    experiment.run_simulations()



"""
This module implements the Connector class, which is an interface to the simulation engine.
Connectors provide a set of functions to interact with the simulation engine,
such as starting and stopping the simulation, adding and removing entities from the simulation,
supporting event generation and management.
"""
import os
from abc import abstractmethod
import abc
import pandas as pd


class Connector(abc.ABC):
    r"""
    This class is an interface to the simulation engine.
    This is an abstract class, and it is meant to be subclassed by connectors to specific simulation engines.

    Parameters
    ----------
    simulation : :class:`~omnetpypy.simulation.Simulation`
        The simulation object that describes the simulation configuration and its context. Used for example to access
        the random number generators for the simulation.
    metrics : list of :class:`~omnetpypy.utilities.FutureMetric` or None, optional
        A list of metrics to be recorded during the simulation.
    output_dir : str, optional
        The output directory where the simulation data will be stored. Default is `None`, which means that the
        simulation data will not be stored.
    repetition : int, optional
        The repetition index of the simulation. Default is 0.


    Attributes
    ----------
    simulation : :class:`~omnetpypy.front_end.simulation.Simulation`
        The simulation object that describes the simulation configuration and its context. Used for example to access
        the random number generators for the simulation.
    metrics : list of :class:`~omnetpypy.utilities.FutureMetric` or None
        A list of metrics to be collected during the simulation. If ``None``, no metrics will be collected.
    output_dir : str
        The output directory where the simulation data will be stored.
    repetition : int
        The repetition index of the simulation.
    """

    def __init__(self, simulation, metrics=None, output_dir=None, repetition=0):
        self.simulation = simulation
        self.metrics = metrics

        self.output_dir = output_dir
        # check if the output directory exists, if not create it
        if output_dir is not None and not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.repetition = repetition

        # allocate a dict of numpy dataframes indexed by metric name
        if metrics is not None:
            metrics_numbers_str = [metric for metric in metrics if metric.type == "number" or metric.type == "str"]
            metrics_dict = [metric for metric in metrics if metric.type == "dict"]
            self.metrics_dict = metrics_dict
            self.metrics_data = {metric.name: pd.DataFrame(columns=["sample", "timestamp"]) for metric in metrics_numbers_str}
            self.metrics_columns = {metric.name: ["sample", "timestamp"] for metric in metrics_numbers_str}
            for metric in metrics_dict:
                self.metrics_data[metric.name] = pd.DataFrame(columns=metric.columns)
                self.metrics_columns[metric.name] = metric.columns

            self.metrics_headers = {metric.name: False for metric in metrics}
        else:
            self.metrics_data = {}
            self.metrics_dict = []
            self.metrics_columns = {}

    @abstractmethod
    def start_simulation(self, until=None):
        """
        Start the simulation.

        Parameters
        ----------
        until : float or None, optional
            The simulation time at which the simulation should stop.
            If None, the simulation will run until there are no more events to process.
        """
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def add_entity(self, entity):
        r"""
        Add an entity to the simulation.

        Parameters
        ----------
        entity : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
            The simulation entity to be added to the simulation.
        """
        entity.set_sim_context(self.simulation)

        # check if entity has property sub_modules
        if hasattr(entity, "sub_modules"):
            for _, entity in entity.sub_modules.items():
                self.add_entity(entity)

    @abstractmethod
    def get_time(self):
        r"""
        Return the current simulation time. The time unit is determined by the simulation configuration.
        """
        raise NotImplementedError("to be implemented by subclasses")

    def random(self):
        r"""
        Return the random number generator instantiated for the current
        simulation.

        Returns
        -------
        :class:`~omnetpypy.utilities.MultiRandom`
            A random number generator that supports multiple independent streams, powered by an independent seed
            for each stream.
        """
        return self.simulation.rng

    @abstractmethod
    def schedule_port_input(self, port, message):
        r"""
        Schedule the event of a message received by a port.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The port that receives the message.
        message : :class:`~omnetpypy.front_end.message.Message`
            The message that received by the port.
        """
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def schedule_self_message(self, message, entity, at=None, delay=None):
        r"""
        Schedule a self message to be processed by an entity.

        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be processed.
        entity : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
            The entity that will process the message.
        at : float or None, optional
            The simulation time at which the message should be processed.
            If None, the ``delay`` parameter will be used.
        delay : float or None, optional
            The time delay from the current simulation time at which the message should be processed.
            If None, the ``at`` parameter will be used.
        """
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def is_scheduled(self, message, entity):
        r"""
        Check if a message is scheduled as a self message for an entity.

        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be checked.
        entity : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
            The entity to check, that should receive the self message.
        """
        raise NotImplementedError("to be implemented by subclasses")

    @abstractmethod
    def cancel_scheduled(self, message, entity):
        r"""
        Cancel a scheduled self message for an entity.
        If the message is not scheduled, nothing happens.
        """
        raise NotImplementedError("to be implemented by subclasses")

    def record_metric(self, metric, value):
        """
        Record a metric value for the current simulation.

        Parameters
        ----------
        metric : str
            The name of the metric to be recorded.
        value : float
            The value of the metric to be recorded.
            If value is a dict or a list and the output file format is csv,
            it will be turned into a string and stored as is under the "sample" column.
        """
        if self.metrics is not None and self.output_dir is not None:
            # add the value to the metric dataframe by simply appending a new row using loc

            if metric in self.metrics_data:

                index_dict = -1
                for i, metric_dict in enumerate(self.metrics_dict):
                    if metric_dict.name == metric:
                        index_dict = i
                        break
                is_dict = True if index_dict != -1 else False

                if not is_dict:
                    self.metrics_data[metric].loc[len(self.metrics_data[metric].index)] = [value, self.get_time()]

                else:
                    assert isinstance(value, dict), "The value of a dict metric must be a dict"
                    assert set(value.keys()) == set(self.metrics_dict[index_dict].columns).difference({"timestamp"}), \
                        "The keys of the dict value must match the columns of the metric"
                    assert "timestamp" not in value.keys(), "The key 'timestamp' is reserved for the timestamp"
                    assert all([isinstance(value[key], (int, float, str)) for key in value.keys()]), \
                        "The values of the dict must be int, float or str"

                    # get the row as a list of values whose keys are ordered alphabetically
                    value["timestamp"] = self.get_time()

                    self.metrics_data[metric].loc[len(self.metrics_data[metric].index)] = value

                # append the metrics to the output file if the dataframes are too big
                if len(self.metrics_data[metric].index) > 1000:
                    self.dump_metric(metric)

        elif self.metrics is None:
            raise Exception("No metrics have been defined for this simulation")

    def dump_metric(self, metric):
        """
        Dump the metric data to the temporary output file for this repetition.

        Parameters
        ----------
        metric : str
            The name of the metric to be dumped.
        """
        if self.metrics is not None and self.output_dir is not None:
            if metric in self.metrics_data:
                if not self.metrics_headers[metric]:
                    self.metrics_data[metric].to_csv(f"{self.output_dir}/.{metric}_vector_rep{self.repetition}.csv", mode="w", header=True, index=False)
                    self.metrics_headers[metric] = True
                else:
                    self.metrics_data[metric].to_csv(f"{self.output_dir}/.{metric}_vector_rep{self.repetition}.csv", mode="a", header=False, index=False)

                self.metrics_data[metric] = pd.DataFrame(columns=self.metrics_columns[metric])

        elif self.metrics is None:
            raise Exception("No metrics have been defined for this simulation")




from omnetpypy import SimpleModule, Message
from omnetpypy import sim_log

__all__ = ["PingPongModule"]

class PingPongModule(SimpleModule):
    """
    The Ping Pong Module. The behavior between the Ping and Pong modules is the same except for
    who sends the first message (Ping). This differentiation is handled in the initialize method through the name.

    Parameters
    ----------
    name : str
        The name of the module. Either "ping" or "pong".
    identifier : int
        The identifier of the module. This identifier should be unique within the simulation.
    delay : int, optional
        The delay in seconds between receiving and sending a message. The default is 5s.
    """

    def __init__(self, name, identifier, delay=5):
        port_names = ["in_out"]
        self.delay = delay
        super().__init__(name, identifier, port_names)

    def initialize(self, step=0):
        if step == 0:
            if self.name == "ping":
                sim_log.info(message="I am the Pinger, generating the first message!", time=self.sim_context.time())

                message = Message(fields=["PING PONG"], header="PING PONG MSG")

                self.schedule_message(message, delay=self.delay)

    def handle_message(self, message, port_name):

        if port_name is None: # self message
            sim_log.info(message=f"{self.name}: Sending the first Message!", time=self.sim_context.time())
            self.send(message, port_name="in_out")
        else:
            sim_log.info(message=f"{self.name}: Received the Message: {message.fields[0]}", time=self.sim_context.time())
            self.send(message, port_name="in_out")
            self.emit_metric("Throughput", 1)



r"""This module provides the tools to parse the YAML configuration and topology files"""
import importlib

import yaml

from omnetpypy.front_end import CompoundModule
from omnetpypy.front_end.channel import Channel


def parse_simple_modules(simple_descriptors):
    r"""
    Parse the simple modules from raw parsed data

    Parameters
    ----------
    simple_descriptors : list
        A list containing the simple modules' data as dictionaries with two keys "name" and "package"

    Returns
    -------
    list of :class:`~omnetpypy.front_end.simple_module.SimpleModule`
        A list of the simple module classes
    """

    # sanity check. Check that the dict has only one key "simple", the value of which is a list,
    # and each element of the list is a dict containing only the keys "name" and "package"
    for entry_dict in simple_descriptors:
        if (not isinstance(entry_dict, dict) or len(entry_dict) != 2 or
                "name" not in entry_dict or "package" not in entry_dict):
            raise ValueError("Invalid simple module format. Each module should have only the keys 'name' and 'package'")

    module_classes = []
    for module_data in simple_descriptors:
        name = module_data['name']
        package = module_data.get('package', None)
        # Import module dynamically
        # import with importlib
        imported = importlib.import_module(name=package)
        module_class = getattr(imported, name)

        module_classes.append(module_class)
    return module_classes


def parse_channels(channel_descriptors):
    r"""
    Parse the channels from raw parsed data

    Parameters
    ----------
    channel_descriptors : list
        A list of dictionaries containing the channel data

    Returns
    -------
    list of :class:`~omnetpypy.front_end.channel.Channel`
        A list of the channel classes
    """

    # sanity check. Check that the dict has only one key "simple", the value of which is a list,
    # and each element of the list is a dict containing only the keys "name" and "package"
    for entry_dict in channel_descriptors:
        if (not isinstance(entry_dict, dict) or len(entry_dict) != 2 or
                "name" not in entry_dict or "package" not in entry_dict):
            raise ValueError("Invalid channel format. Each module should have only the keys 'name' and 'package'")

    channel_classes = []
    for channel_descriptors in channel_descriptors:
        name = channel_descriptors['name']
        package = channel_descriptors['package']
        # Import channel dynamically if not already imported
        imported = importlib.import_module(name=package)
        channel_class = getattr(imported, name)
        channel_classes.append(channel_class)
    return channel_classes


def parse_connection(connection_data, compound_module, channel_classes, next_channel_id):
    r"""
    Parse a connection from raw parsed data and entity classes within a compound module

    Parameters
    ----------
    connection_data : dict
        A dictionary containing the connection data
    compound_module : :class:`~omnetpypy.front_end.compound_module.CompoundModule`
        The compound module instance
    channel_classes : list
        A list of the channel classes, already imported
    next_channel_id : int
        The current next channel identifier, used to generate unique identifiers for channels

    Returns
    -------
    int
        The new next channel identifier

    Raises
    ------
    Exception
        If the connection data is invalid
    """
    source = connection_data['source']
    target = connection_data['target']
    source_module, source_port = source.split('.')
    target_module, target_port = target.split('.')
    if source_module == "self":
        source_module = compound_module
    else:
        source_module = compound_module.sub_modules[source_module]
    if target_module == "self":
        target_module = compound_module
    else:
        target_module = compound_module.sub_modules[target_module]

    # check if the "type" field is present

    if "type" in connection_data and connection_data["type"] == "forward input":
        source_module.ports[source_port].forward_input(target_module.ports[target_port])

    elif "type" in connection_data and connection_data["type"] == "forward output":
        source_module.ports[source_port].forward_output(target_module.ports[target_port])

    elif "type" in connection_data and connection_data["type"] == "subscribed":
        target_module.ports[target_port].subscribe_to(source_module.ports[source_port])

    elif "type" not in connection_data and "channel" not in connection_data:
        channel = None
        source_module.connect(local_port=source_port, remote_entity=target_module, remote_port=target_port,
                              channel=channel)

    elif "type" not in connection_data and "channel" in connection_data:
        if connection_data["channel"] == "default":
            # in this case we gather the parameters and instantiate a default channel
            parameters = connection_data.get("parameters", {}).copy()
            parameters["name"] = "channel_" + str(next_channel_id)
            parameters["identifier"] = next_channel_id
            next_channel_id += 1
            channel = Channel(**parameters)
        else:
            # in this case we have to instantiate the channel
            channel_classname = connection_data["channel"]
            # find the channel class iterating over the module_classes and looking for the class with the same name
            channel_class = None
            for channel_class in channel_classes:
                if channel_class.__name__ == channel_classname:
                    channel_class = channel_class
                    break
            parameters = connection_data.get("parameters", {})
            parameters["name"] = f"{channel_class}_" + str(next_channel_id)
            parameters["identifier"] = next_channel_id
            next_channel_id += 1
            channel = channel_class(**parameters)

        # the channel is added to the compound module as a sub entity
        compound_module.add_sub_module(channel)

        source_module.connect(local_port=source_port, remote_entity=target_module, remote_port=target_port,
                              channel=channel)
    else:
        raise Exception(f"Invalid connection: {connection_data}")

    return next_channel_id


def sanitize_compound_descriptors(compound_descriptors):
    r"""
    Sanitize the compound descriptors to ensure there is no submodule dependency loop that would cause endless
    recursion.

    Parameters
    ----------
    compound_descriptors : list
        The list of compound module descriptors

    Raises
    ------
    ValueError
        If a dependency loop is detected or if one of the submodules has an invalid name
    """

    def check_dependencies(compound_descriptor, stack):
        if compound_descriptor['name'] in stack:
            raise ValueError(f"Dependency loop detected in compound module {compound_descriptor['name']}")
        stack.append(compound_descriptor['name'])
        for submodule in compound_descriptor.get('submodules', []):
            type_name = submodule['type']
            submodule_name = submodule['name']
            if submodule_name == "self":
                raise ValueError(f"Invalid submodule name 'self' in compound module {compound_descriptor['name']}")
            for desc in compound_descriptors:
                if desc['name'] == type_name:
                    check_dependencies(desc, stack)
        stack.pop()

    for descriptor in compound_descriptors:
        check_dependencies(descriptor, [])


def instantiate_compound_module(name, module_classes, channel_classes, compound_descriptors,
                                next_module_id, next_channel_id):
    r"""
    Instantiate a compound module from its yaml descriptor. The function is recursive, as a compound module can contain
    other compound modules. In such case, inner compound modules are fully initialized first.

    Parameters
    ----------
    name : str
        The name of the compound module
    module_classes : list
        A list of the simple module classes, already imported
    channel_classes : list
        A list of the channel classes, already imported
    compound_descriptors : list
        The list of compound module descriptors, used to instantiate the compound modules when needed
    next_module_id : int
        The current next simple module identifier
    next_channel_id : int
        The current next channel identifier

    Returns
    -------
    tuple
        The compound module instance (:class:`~omnetpypy.front_end.compound_module.CompoundModule`), the new next simple
        module identifier, the new next channel identifier, and the new next compound module identifier

    Raises
    ------
    ValueError
        If the compound module name is not found in the descriptors or one of its submodules was not defined
    """

    descriptor = None
    for desc in compound_descriptors:
        if desc['name'] == name:
            descriptor = desc
            break
    if descriptor is None:
        raise ValueError(f"Compound module {name} not found in the descriptors")

    ports = [port['name'] for port in descriptor.get('ports', [])]
    parameters = {'name': name, 'identifier': next_module_id, 'port_names': ports, 'parent': None}
    kwargs = descriptor.get('parameters', {})
    parameters.update(kwargs)
    next_module_id += 1
    compound_module = CompoundModule(**parameters)
    for submodule_data in descriptor.get('submodules', []):
        type_name = submodule_data['type']
        # find the submodule class iterating over the module_classes and looking for the class with the same name
        submodule_instance = None
        # first look for a simple module with that name
        for module_class in module_classes:
            if module_class.__name__ == type_name:
                submodule_class = module_class

                sub_name = submodule_data['name']
                parameters = submodule_data.get('parameters', {})
                parameters['name'] = sub_name
                parameters['identifier'] = next_module_id
                next_module_id += 1
                submodule_instance = submodule_class(**parameters)

                break
        # if not found, look for a compound module with that name
        for compound_desc in compound_descriptors:
            if compound_desc['name'] == type_name:

                submodule_instance, next_module_id, next_channel_id = instantiate_compound_module(type_name,
                                                                                                  module_classes,
                                                                                                  channel_classes,
                                                                                                  compound_descriptors,
                                                                                                  next_module_id,
                                                                                                  next_channel_id)

                break

        # if still not found, raise an exception
        if submodule_instance is None:
            raise ValueError(f"No simple or compound module with name {type_name} was found")

        compound_module.add_sub_module(submodule_instance)

    connections = descriptor.get('connections', [])

    for connection_data in connections:
        # we check whether there is a field name "for i in x to y"
        # if so, we have to create a connection for iteration
        # get the keys of the dictionary and check if one starts with "for"
        keys = connection_data.keys()
        loop_found = False
        for key in keys:
            if key.startswith("for"):
                # get the range of the iteration
                bottom = int(key.split(" ")[3])
                top = int(key.split(" ")[5])
                # get the name of the variable
                var_name = key.split(" ")[1]
                # get the connection data
                conn_data = connection_data[key]
                # iterate over the range
                for i in range(bottom, top + 1):
                    # format conn_data['source'] and conn_data['target'] by replacing the variable name
                    new_data = conn_data.copy()
                    for k, v in new_data.items():
                        if k == "source" or k == "target":
                            new_data[k] = v.replace("{" + var_name + "}", str(i))

                            # there might also occur "{i+l}" or "{i-l}" in the string
                            # we have to replace these as well using pattern matching
                            # we can use the re module for this
                            # we have to import the re module
                            import re
                            # we have to find all occurences of "{i+l}" and "{i-l}" in the string, for any l
                            # we can use the findall method of the re module
                            # we have to define the pattern first
                            pattern = r"{i[+-]\d+}"
                            # then we can use the findall method
                            matches = re.findall(pattern, new_data[k])
                            # we have to iterate over the matches and replace them in the string
                            for match in matches:
                                # we have to extract the number and the operator from the match
                                operator = match[2]
                                number = int(match[3:-1])
                                # we have to replace the match in the string
                                if operator == "+":
                                    new_data[k] = new_data[k].replace(match, str(i + number))
                                else:
                                    new_data[k] = new_data[k].replace(match, str(i - number))

                    # parse the connection
                    next_channel_id = parse_connection(new_data, compound_module, channel_classes,
                                                       next_channel_id)
                    loop_found = True

        if not loop_found:
            next_channel_id = parse_connection(connection_data, compound_module, channel_classes,
                                               next_channel_id)

    return compound_module, next_module_id, next_channel_id


def parse_yaml_file(file_path):
    r"""
    Parse the YAML configuration file.

    Parameters
    ----------
    file_path : str
        The path to the YAML configuration file

    Returns
    -------
    dict
        The parsed configuration, as a dictionary directly loaded from the YAML file
    """
    with open(file_path, 'r') as file:
        config = yaml.safe_load(file)
    return config


def parse_yaml_directory(directory_path, sim_context):
    r"""
    Parse all the YAML files from the specified directory. Files are parsed in the order listed below.
    The directory should contain:

        - "config.yaml": the configuration file
        - "simple.yaml": the file listing all the simple modules used in the simulation
        - "channels.yaml" (optional): the file listing all the custom channels used in the simulation
        - "compound.yaml" (optional): the file listing all the compound modules used in the simulation
        - "network.yaml": the network topology file

    Parameters
    ----------
    directory_path : str
        The path to the directory containing the YAML files
    sim_context : :class:`~omnetpypy.simulation.Simulation`
        The simulation with its context

    Returns
    -------
    tuple
        A tuple containing the configuration, the simple modules, the channels, and the compound modules

    Raises
    ------
    FileNotFoundError
        If one of the mandatory files is missing
    ValueError
        If one of the files has an invalid format or there is a dependency loop in the compound modules

    """
    initial_module_id = 10100
    initial_channel_id = 2222200000

    # Parse the simple modules
    simple_data = parse_yaml_file(directory_path + "./simple.yaml")
    # sanity check. Check that the dict has only one key "simple", the value of which is a list,
    # and each element of the list is a dict containing only the keys "name" and "package"
    if len(simple_data) != 1 or not isinstance(simple_data["simple"], list):
        raise ValueError("Invalid simple module format. It should have only one key 'simple' with a list of modules")
    simple_descriptors = simple_data.get("simple", [])
    simple_modules = parse_simple_modules(simple_descriptors)

    # Parse the channels
    # checking if the file exists
    try:
        channel_data = parse_yaml_file(directory_path + "./channels.yaml")
        if len(channel_data) != 1 or not isinstance(channel_data["channels"], list):
            raise ValueError(
                "Invalid channel yaml format. It should have only one key 'channels' with a list of channels")
        channel_data = channel_data.get("channels", [])
    except FileNotFoundError:
        channel_data = []
    channel_classes = parse_channels(channel_data)

    # Parse the compound modules
    # checking if the file exists
    try:
        compound_data = parse_yaml_file(directory_path + "./compound.yaml")
        compound_data = compound_data.get("compound", [])
        sanitize_compound_descriptors(compound_data)
    except FileNotFoundError:
        compound_data = []
    compound_descriptors = compound_data

    # Parse the network topology
    network = parse_yaml_file(directory_path + "./network.yaml")
    # sanity check. Check that the dict has only one key "network", the value of which is a list,
    if not isinstance(network["network"], list) or len(network["network"]) != 1:
        raise ValueError("Invalid network format. It should have only one key 'network'")
    network = network["network"]

    network_name = network[0]['name']

    # concatenate
    compound_descriptors += network

    network_module, _, _ = instantiate_compound_module(network_name, simple_modules, channel_classes,
                                                       compound_descriptors, initial_module_id, initial_channel_id)
    sim_context.connector.add_entity(network_module)

    return network_module



"""
This module implements the PI controller for the AQM algorithm.
"""
from omnetpypy import sim_log


class PIController:

    def __init__(self):
        self.alpha = None
        self.beta = None
        self.q_ref = None

        self.q_old = 0
        self.p_old = 0

        self.q = 0
        self.p = 0

    def update(self, q):
        self.q = q
        self.p = self.alpha * (self.q - self.q_ref) - self.beta * (self.q_old - self.q_ref) + self.p_old
        # print(f"q_ref={self.q_ref}, q={self.q}, p={self.p}, p_old={self.p_old}, q_old={self.q_old}")
        self.q_old = self.q
        self.p_old = self.p

    def get_marking_probability(self):
        return self.p
        # return 0.

    def set_parameters(self, R_plus, C, N_minus, q_ref):
        """
        Set the a,b parameters for the PI controller from R_plus, C, N_minus, T

        Parameters
        ----------
        R_plus : float
            The maximum Round Trip Time (RTT) in seconds for which the PI controller is stable
        C : float
            The channel capacity in LLE attempts per second
        N_minus : float
            The minimum number of TCP-like flows passing through the link
        q_ref : float
            The reference queue size in (avg) LLE attempts

        Returns
        -------
        T : float
            The time interval in seconds for the PI controller sampling
        """

        omega_g = 2*N_minus/(R_plus*R_plus*C)
        assert omega_g < 0.05/R_plus, "The PI controller could not be stable"
        p_queue = 1/R_plus

        T = 1/(omega_g*100)

        K_PI = (complex(0, omega_g)/p_queue + 1)/((R_plus*C)**3/(2*N_minus)**2)
        # turn K_PI into a real number by taking the module
        K_PI = abs(K_PI)*omega_g*100  # TODO: remove the *10

        assert 1-omega_g*T > 0, "The PI controller could not be stable"
        self.alpha = K_PI/omega_g
        self.beta = self.alpha*(1-omega_g*T)
        self.q_ref = q_ref

        sim_log.debug(f"PI controller parameters: alpha={self.alpha}, beta={self.beta}, q_ref={q_ref}, K_PI={K_PI},"
                      f"omega_g={omega_g}, T={T}")

        return T



"""
Read the file "queue_size_vector.csv" and plot the queue size as a function of time
"""

import pandas as pd
import matplotlib.pyplot as plt

if __name__ == "__main__":
    df = pd.read_csv("./out/queue_size_vector.csv")

    fig, axs = plt.subplots(2, sharex=True)
    ax = axs[0]
    ax2 = axs[1]

    ax2.set_xlabel("Time (ms)")

    # now we want to plot as a sliding window average
    window_size_time_units = 10000
    # every point in the plot is the average of all samples in the window
    # we can't use directly group by because the same sample may be in multiple adjacent windows
    # so, for every row in df, we compute the average of all samples in the window
    # we use a list comprehension to do that
    sw_avg = pd.Series([df[(df["timestamp"] >= t - window_size_time_units) & (df["timestamp"] < t)]["sample"].mean() for t in df["timestamp"]])

    df["timestamp"] = df["timestamp"] / 1e3  # convert to ms

    # multiply the sample column by 0.04
    # sw_avg *= 0.04

    ax.plot(df["timestamp"], sw_avg, label="queue size", color="purple")

    ax.set_ylabel("Queue Size (# requests)", color="purple")

    # also plot a horizontal green dashed line at the reference queue size set to 20
    ax.axhline(y=10, color='magenta', linestyle='--', label="reference queue size")

    cur_flows = 4
    delete_phase = False
    # create a colormap for the background
    colors = {4: "blue", 10: "yellow", 16: "red"}
    for i in range(0, 48000, 8000):
        ax.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        ax2.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        # add some text next to the line saying "2 flows added" at the top
        ax.text(i + 1000, 45, f"{cur_flows} flows", rotation=0, fontsize=10)

        # add a colored background
        ax.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)
        ax2.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)

        if cur_flows == 16:
            delete_phase = True
        if cur_flows == 4:
            delete_phase = False
        if delete_phase:
            cur_flows -= 6
        else:
            cur_flows += 6

    # read the IRG file and plot the Inter-Request Gap (IRG) on the same plot but on a different axis
    df_irg = pd.read_csv("./out/IRG_vector.csv")
    window_size_time_units = 50000
    sw_avg_irg = pd.Series([df_irg[(df_irg["timestamp"] >= t - window_size_time_units) & (df_irg["timestamp"] < t)]["sample"].mean() for t in df_irg["timestamp"]])

    df_irg["timestamp"] /= 1e3

    # set y axis in ms
    sw_avg_irg /= 1e3

    ax2.plot(df_irg["timestamp"], sw_avg_irg, color="green", label="IRG (flow 0)")
    # ax2.set_ylabel("Inter-Request Gap (IRG) (ms)", color="blue")
    ax2.set_ylabel("Inter-Request Gap (ms)", color="green")

    # set the maximum value for the y-axis
    # ax2.set_ylim(0, 16)

    ax.set_title("Queue size and IRG (flow 0)")

    lines, labels = ax.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()

    ax.legend(lines, labels, loc="upper right")
    # ax2.legend(lines2, labels2, loc="lower right")

    plt.tight_layout()

    plt.show()

    # Save the figure
    fig.savefig("./out/queue_size.pdf")




r"""
The omnetpupy backends sub-package is responsible for the connection between the user API
and discrete event simulation engines.

The user API is defined in the front end sub-package, and is independent of the simulation engine used.
The backends sub-package provides the necessary tools to connect the user API to the simulation engine,
and to run the simulation.
"""

from omnetpypy.backends.connector import Connector

__all__ = ["Connector"]



"""
plot the congestion window from ./out/congestion_window_vector.csv as a function of time (ms)
there is only one flow
"""
import math

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

def compute_fid(latency, gamma):
    """
    Compute the flow id from the latency and the gamma parameter
    """
    return 0.5 + 0.5*math.exp(-2*gamma*latency)

if __name__ == "__main__":

    fig, axs = plt.subplots(2, sharex=True)
    ax = axs[0]
    ax2 = axs[1]
    """
    ax.plot(df["timestamp"], df["sample"], label="congestion window")
    """

    ax2.set_xlabel("Time (ms)")
    # ax.set_ylabel("Congestion Window Size")
    ax.set_ylabel("Latency (ms)", color="red")
    ax.set_title("Latency and Throughput")
    ax.set_ylim(0, 40)

    # make the y-axis labels be at the same x position
    ax.yaxis.set_label_coords(x=-0.06, y=0.5)
    ax2.yaxis.set_label_coords(x=-0.06, y=0.5)

    # read also the file "latency_vector.csv" and plot a moving average of latency as a function of time in the same plot
    df_lat = pd.read_csv("./out/latency_vector.csv")

    # Compute the sliding window average
    window_size_time_units = 5000

    # create a new column with the sliding window index
    df_lat["sw_index"] = df_lat["timestamp"] // window_size_time_units

    # compute the sliding window average
    sw_avg = pd.Series(
        [df_lat[(df_lat["timestamp"] >= t - window_size_time_units) & (df_lat["timestamp"] < t)]["sample"].mean() for t
         in df_lat["timestamp"]])

    # divide the timestamp by 1e3 to have ms
    df_lat["timestamp"] /= 1e3
    sw_avg /= 1e3

    # ax2 = ax.twinx()
    ax.plot(df_lat["timestamp"], sw_avg, color="red", label="latency")
    # ax.set_ylabel("Latency (ms)")

    # still on ax2 print the average secret key rate for the second half of the simulation
    ax.axhline(y=sw_avg[sw_avg.size // 2:].mean(), color='red', linestyle='--',
                label="AVG latency at steady state")

    """cur_flows = 4
    delete_phase = False
    # create a colormap for the background
    colors = {4: "blue", 10: "yellow", 16: "red"}
    for i in range(0, 48000, 8000):
        ax.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        ax2.axvline(x=i, color='black', linestyle=':', alpha=0.5)
        # add some text next to the line saying "2 flows added" at the top
        ax.text(i + 1000, 95, f"{cur_flows} flows", rotation=0, fontsize=10)

        # add a colored background
        ax.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)
        ax2.axvspan(i, i + 8000, color=colors[cur_flows], alpha=0.1)

        if cur_flows == 16:
            delete_phase = True
        if cur_flows == 4:
            delete_phase = False
        if delete_phase:
            cur_flows -= 6
        else:
            cur_flows += 6"""


    # read the IRG file and plot the Inter-Request Gap (IRG) on the same plot but on a different axis
    df_irg = pd.read_csv("./out/throughput_vector.csv")
    window_size_time_units *= 5
    # sw_avg_irg = pd.Series([df_irg[(df_irg["timestamp"] >= t - window_size_time_units) & (df_irg["timestamp"] < t)]["sample"].mean() for t in df_irg["timestamp"]])
    sw_avg_irg = pd.Series([df_irg[(df_irg["timestamp"] >= t - window_size_time_units) & (df_irg["timestamp"] < t)]["sample"].count() for t in df_irg["timestamp"]])
    sw_avg_irg /= window_size_time_units # pairs per us
    sw_avg_irg *= 1e3  # pairs per ms

    df_irg["timestamp"] /= 1e3

    ax2.plot(df_irg["timestamp"], sw_avg_irg, color="blue", label="E2E Throughput")
    # ax2.set_ylabel("Inter-Request Gap (IRG) (ms)", color="blue")
    ax2.set_ylabel("Throughput (pairs/ms)", color="blue")
    # ax2.set_ylim(0, 4)

    # still on ax2 print the average secret key rate for the second half of the simulation
    ax2.axhline(y=sw_avg_irg[sw_avg_irg.size // 2:].mean(), color='blue', linestyle='--',
                label="AVG E2E SKR at steady state")

    """
    # we add a third axis to plot the throughput, again as a moving average
    throughput_window_size_time_units = 1000
    ax3 = ax.twinx()
    df_throughput = pd.read_csv("./out/throughput_vector.csv")
    df_throughput["sw_index"] = df_throughput["timestamp"] // throughput_window_size_time_units
    sw_avg_throughput = df_throughput.groupby("sw_index").count()
    sw_avg_throughput["sample"] /= throughput_window_size_time_units  # pairs per ms

    ax3.plot(sw_avg_throughput.index, sw_avg_throughput["sample"], color="green", label="throughput")
    """
    """
    df_fid = sw_avg
    df_fid["fid"] = df_fid.apply(lambda row: compute_fid(row["sample"], 0.01), axis=1)
    ax2 = ax.twinx()
    ax2.plot(df_fid["timestamp"], df_fid["fid"], color="green", label="Fidelity")
    ax2.set_ylabel("Fidelity")
    """


    # print the legends within the axis

    lines, labels = ax.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    # lines3, labels3 = ax3.get_legend_handles_labels()
    # ax2.legend(lines + lines2, labels + labels2, loc="upper right")

    # print legend in the axis box
    # ax.legend(lines, labels, loc="best")
    # ax2.legend(lines2, labels2, loc="best")

    # adjust plot size
    fig.tight_layout()

    # save the plot
    plt.savefig("./out/latency_dynamic.pdf")

    plt.show()




"""
This toy simulation module implements a simple ping-pong protocol.
The simulation consists of two modules, a `PingModule` and a `PongModule`,
which send messages to each other in a ping-pong fashion. The `PingModule` sends a message to the `PongModule`,
which then sends a message back to the `PingModule`. This process is repeated for a specified number of times.
"""

from omnetpypy.examples.ping_pong.ping_pong_simple import PingPongModule

__all__ = ["PingPongModule"]



"""
This module implements the :class:`~omnetpypy.front_end.message.Message` class,
representing a message in the simulation.
"""

__all__ = ['Message']


class Message:
    r"""
    This class is a wrapper for messages exchanged between entities in a simulation.

    Parameters
    ----------
    fields : list
        The fields of the message.
    meta : dict, optional
        Additional metadata to be stored with the message. A typical use case is to store a "header".

    Attributes
    ----------
    fields : list
        The fields of the message.
    meta : dict
        Additional, editable metadata stored with the message.
    """

    def __init__(self, fields, **meta):
        self.fields = fields
        self.meta = meta
        if meta is None:
            self.meta = {}

    def __str__(self):
        return f"Message(fields={self.fields}, meta={self.meta})"

    def __copy__(self):
        """
        Return a shallow copy of the message.
        """
        new_meta = self.meta.copy()
        return Message(fields=self.fields[:], **new_meta)



import abc

from omnetpypy import sim_log


class AIMDCongestionController(abc.ABC):
    """
    A simple congestion controller that uses the Additive Increase Multiplicative Decrease (AIMD) algorithm.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def halve_congestion_knob(self, flow_id, current_time=None):
        """
        Halve the congestion metric for the given flow
        """
        pass

    @abc.abstractmethod
    def increase_congestion_knob(self, flow_id, current_time=None):
        """
        Increase the congestion metric for the given flow
        """
        pass

    @abc.abstractmethod
    def setup_congestion_control(self, flow, current_time=None):
        """
        Setup the congestion control variables for the given flow
        """
        pass

    @abc.abstractmethod
    def delete_flow(self, flow_id):
        """
        Delete the flow from the congestion controller
        """
        pass

    @abc.abstractmethod
    def collect_timeouts(self, current_time):
        """
        Collect the timeouts for the requests in flight and update the congestion metric for the involved flows
        """
        pass

    @abc.abstractmethod
    def handle_ack(self, flow_id, req_id, current_time, time_sent, mark_congested=False):
        """
        Handle the ACK for the given request

        Parameters
        ----------
        flow_id : int
            The flow id
        req_id : int
            The request id
        current_time : float
            The current time
        time_sent : float
            The time the request was sent
        mark_congested : bool, optional
            Whether the ack was marked as congested or not

        Returns
        -------
        int
            The number of new requests to generate for the flow
        """
        pass

    @abc.abstractmethod
    def handle_new_request_in_flight(self, flow_id, req_id, current_time):
        """
        Handle a new request in flight for the given flow
        """
        pass


class WindowCongestionController(AIMDCongestionController):
    """
    A simple congestion controller that uses the Additive Increase Multiplicative Decrease (AIMD) algorithm
    with a window-based approach.
    """
    def __init__(self):

        self.congestion_windows = {}  # dictionary of congestion windows, one for each flow, indexed by flow_id
        self.ssthresh = {}  # dictionary of slow start thresholds, one for each flow, indexed by flow_id
        self.is_slow_start = {}  # dictionary of flags indicating if the flow is in slow start, indexed by flow_id
        self.estimated_rtt = {}
        self.dev_rtt = {}
        self.requests_in_flight = {}  # dictionary of requests in flight, one list for each flow, indexed by flow_id
        # every element in the list is a tuple (req_id, time_sent, timeout)
        self.max_congestion_window = 1000  # maximum congestion window size
        self.consecutive_acks = {}  # dictionary of consecutive acks, one for each flow, indexed by flow_id
        self.consecutive_acks_required = 1  # number of consecutive acks to receive before increasing the congestion window

        self.other_ends = {}  # dictionary of other end identifiers, one for each flow, indexed by flow_id

        super().__init__()

    def halve_congestion_knob(self, flow_id, current_time=None):
        """
        Halve the congestion window for the given flow
        """

        self.ssthresh[flow_id] = max(self.congestion_windows[flow_id] / 2, 1)
        self.congestion_windows[flow_id] = 1.
        self.is_slow_start[flow_id] = True

    def increase_congestion_knob(self, flow_id, current_time=None):
        """
        Increase the congestion window for the given flow
        """
        # sim_log.info(f"Flow {flow_id} congestion window increased to {self.congestion_windows[flow_id]}")
        if self.is_slow_start[flow_id]:
            # if the flow is in slow start, double the congestion window
            self.congestion_windows[flow_id] = min(self.max_congestion_window, self.congestion_windows[flow_id] + 1)
            if self.congestion_windows[flow_id] >= self.ssthresh[flow_id]:
                # if the congestion window is greater than the slow start threshold, we are not in slow start anymore
                self.is_slow_start[flow_id] = False
            # sim_log.error(f"Flow {flow_id} congestion window increased to {self.congestion_windows[flow_id]}")
        else:
            # if the flow is not in slow start, increase the congestion window linearly
            increment = 1/self.congestion_windows[flow_id]
            old_congestion_window = self.congestion_windows[flow_id]
            self.congestion_windows[flow_id] = min(self.max_congestion_window, self.congestion_windows[flow_id] + increment)
            # sim_log.info(f"Flow {flow_id} congestion window increased from {old_congestion_window} to {self.congestion_windows[flow_id]}")

    def setup_congestion_control(self, flow, current_time=None):
        """
        Setup the congestion control variables for the given flow
        """
        self.congestion_windows[flow["flow_id"]] = 1.  # initial congestion window size
        self.estimated_rtt[flow["flow_id"]] = 300 * (len(flow["path"]) - 1) * 10  # 3000 us per hop is the initial value
        self.dev_rtt[flow["flow_id"]] = 0.05 * self.estimated_rtt[flow["flow_id"]]  # 5% of the estimated RTT
        self.requests_in_flight[flow["flow_id"]] = []
        self.other_ends[flow["flow_id"]] = flow["destination"]
        self.consecutive_acks[flow["flow_id"]] = 0  # number of consecutive acks to receive before increasing the congestion window
        self.ssthresh[flow["flow_id"]] = float("inf")  # slow start threshold
        self.is_slow_start[flow["flow_id"]] = True  # flag indicating if the flow is in slow start

    def delete_flow(self, flow_id):
        """
        Delete the flow from the congestion controller
        """
        del self.congestion_windows[flow_id]
        del self.estimated_rtt[flow_id]
        del self.dev_rtt[flow_id]
        del self.requests_in_flight[flow_id]
        del self.other_ends[flow_id]
        del self.consecutive_acks[flow_id]
        del self.ssthresh[flow_id]
        del self.is_slow_start[flow_id]

    def collect_timeouts(self, current_time):
        """
        Collect the timeouts for the requests in flight and update the congestion windows for the involved flows
        """
        for flow_id in self.requests_in_flight:
            timed_out = False
            requests = self.requests_in_flight[flow_id]
            new_requests = []
            for req_id, time_sent, timeout in requests:
                if current_time - time_sent > timeout:
                    # the request has timed out
                    # we have to halve the congestion window
                    timed_out = True
                else:
                    new_requests.append((req_id, time_sent, timeout))
            self.requests_in_flight[flow_id] = new_requests
            if timed_out:
                """
                sim_log.info(
                f"Flow {flow_id} timed out {len(requests) - len(new_requests)} times, halving congestion window {self.congestion_windows[flow_id]}")
                """
                self.halve_congestion_knob(flow_id)
                pass

    def handle_ack(self, flow_id, req_id, current_time, time_sent, mark_congested=False):
        """
        Handle the ACK for the given request

        Parameters
        ----------
        flow_id : int
            The flow id
        req_id : int
            The request id
        current_time : float
            The current time
        time_sent : float
            The time the request was sent
        mark_congested : bool, optional
            Whether the ack was marked as congested or not

        Returns
        -------
        int
            The number of new requests to generate for the flow
        """
        num_skipped = 0

        found = False

        # update the estimated RTT
        sample_rtt = current_time - time_sent
        self.estimated_rtt[flow_id] = 0.875 * self.estimated_rtt[flow_id] + 0.125 * sample_rtt
        self.dev_rtt[flow_id] = 0.75 * self.dev_rtt[flow_id] + 0.25 * abs(sample_rtt - self.estimated_rtt[flow_id])
        # print(f"Estimated RTT for flow {flow_id} is {self.estimated_rtt[flow_id]}, dev RTT is {self.dev_rtt[flow_id]}")

        for req, time_sent, timeout in self.requests_in_flight[flow_id]:
            if req < req_id:
                num_skipped += 1
                # if we receive an ACK for a request, we can remove all the requests with lower req_id
                # because they have been surely dropped due to congestion
            elif req == req_id:
                # we found the request for which we received the ACK
                # remove it from the list
                found = True

        if found:
            # remove the tuples from the list
            self.requests_in_flight[flow_id] = self.requests_in_flight[flow_id][num_skipped + 1:]

        else:
            self.requests_in_flight[flow_id] = self.requests_in_flight[flow_id][num_skipped:]
            """
            print(f"Flow {flow_id} received an ACK for a request that was not in flight, num_skipped={num_skipped}, "
                  f"req_id={req_id}, requests_in_flight={self.requests_in_flight[flow_id]}")
            """

        if mark_congested:
            sim_log.warning(f"Flow {flow_id} marked as congested")
            self.halve_congestion_knob(flow_id)

        elif num_skipped > 0:
            # sim_log.error(f"Flow {flow_id} marked as congested due to losses")
            self.halve_congestion_knob(flow_id)

        if num_skipped > 0 or mark_congested:
            self.consecutive_acks[flow_id] = 0

        # we increase the congestion window
        if found:
            self.consecutive_acks[flow_id] += 1
        if self.consecutive_acks[flow_id] == self.consecutive_acks_required:
            self.increase_congestion_knob(flow_id)
            self.consecutive_acks[flow_id] = 0

        # return the number of new requests to generate
        return max(int(self.congestion_windows[flow_id]) - len(self.requests_in_flight[flow_id]), 0)

    def handle_new_request_in_flight(self, flow_id, req_id, current_time):
        """
        Handle a new request in flight for the given flow
        """
        timeout = max(self.estimated_rtt[flow_id] + 4 * self.dev_rtt[flow_id], 0.1)
        self.requests_in_flight[flow_id].append((req_id, current_time, timeout))

    def get_congestion_window(self, flow_id):
        """
        Get the congestion window for the given flow

        Returns
        -------
        int
            The congestion window size
        """
        return self.congestion_windows[flow_id]


class RateCongestionController(AIMDCongestionController):
    """
    A simple congestion controller that uses the Additive Increase Multiplicative Decrease (AIMD) algorithm
    with a rate-based approach. The knob is the new request generation probability, which is used as input
    to a geometric process to generate new requests.

    Additive Increase: increase the rate with a constant slope over time
    Multiplicative Decrease: halve the rate when a congestion event is detected
    """

    def __init__(self, C=50000000):
        self.congestion_knobs = {}  # dictionary of congestion knobs, one for each flow, indexed by flow_id
        self.ssthresh = {}  # dictionary of slow start thresholds, one for each flow, indexed by flow_id
        self.is_slow_start = {}  # dictionary of flags indicating if the flow is in slow start, indexed by flow_id
        self.last_update = {}  # dictionary of last update times, one for each flow, indexed by flow_id
        self.last_halved = {}  # dictionary of last halved times, one for each flow, indexed by flow_id
        self.other_ends = {}  # dictionary of other end identifiers, one for each flow, indexed by flow_id
        self.C = {}  # dictionary of C values, one for each flow, indexed by flow_id
        self.default_C = C  # default C value

        self.initial_congestion_knob = 48000  # initial congestion knob (IRG), (us)
        self.max_ssthresh = 1024000  # maximum slow start threshold

        self.estimated_rtt = {}
        self.dev_rtt = {}
        self.requests_in_flight = {}  # dictionary of requests in flight, one list for each flow, indexed by flow_id
        # every element in the list is a tuple (req_id, time_sent, timeout)

        super().__init__()

    def halve_congestion_knob(self, flow_id, current_time=None):
        """
        Halve the congestion knob (rate in this case) for the given flow
        """

        # if last half was less than 3RTT ago, do not halve again
        if current_time - self.last_halved[flow_id] < 3*self.estimated_rtt[flow_id]:
            sim_log.debug(f"Flow {flow_id} congestion knob halved ({current_time - self.last_update[flow_id]}) less than an RTT ({self.estimated_rtt[flow_id]}) ago ), not halving again")
            return

        self.ssthresh[flow_id] = min(self.congestion_knobs[flow_id] * 4, self.max_ssthresh)
        self.congestion_knobs[flow_id] = max(self.initial_congestion_knob,
                                             self.ssthresh[flow_id] * 4)
        self.last_update[flow_id] = current_time
        self.last_halved[flow_id] = current_time
        self.is_slow_start[flow_id] = True

        sim_log.info(f"Flow {flow_id} congestion knob halved to {self.congestion_knobs[flow_id]} with ss thresh {self.ssthresh[flow_id]}")
        if current_time is None:
            raise ValueError("Current time must be provided with the rate based controller")

    def increase_congestion_knob(self, flow_id, current_time=None):
        """
        Increase the congestion knob for the given flow
        """
        if current_time is None:
            raise ValueError("Current time must be provided with the rate based controller")
        # elapsed_time = current_time - self.last_update[flow_id]  # us
        if flow_id not in self.is_slow_start:
            raise ValueError(f"Flow {flow_id} not in the congestion controller")
        if not self.is_slow_start[flow_id]:
            new_IRG = (self.C[flow_id] * self.congestion_knobs[flow_id]) / (self.C[flow_id] + self.congestion_knobs[flow_id])
            self.congestion_knobs[flow_id] = new_IRG
        else:
            new_IRG = self.congestion_knobs[flow_id] / 1.1
            self.congestion_knobs[flow_id] = new_IRG
            if new_IRG <= self.ssthresh[flow_id]:
                # print(self.ssthresh[flow_id])
                self.is_slow_start[flow_id] = False

        self.last_update[flow_id] = current_time

        # print(f"Flow {flow_id} congestion knob increased to {self.congestion_knobs[flow_id]}")
        # print(f"Flow {flow_id} congestion knob increased by {increase}")

    def setup_congestion_control(self, flow, current_time=None, is_source=True):
        """
        Setup the congestion control variables for the given flow
        """
        if current_time is None:
            raise ValueError("Current time must be provided with the rate based controller")
        self.congestion_knobs[flow["flow_id"]] = self.initial_congestion_knob  # initial congestion knob (IPG), (us)
        self.last_update[flow["flow_id"]] = current_time
        self.last_halved[flow["flow_id"]] = current_time
        self.other_ends[flow["flow_id"]] = flow["destination"] if is_source else flow["source"]
        self.ssthresh[flow["flow_id"]] = 1200.  # initial slow start threshold
        self.is_slow_start[flow["flow_id"]] = True

        self.estimated_rtt[flow["flow_id"]] = 300 * (len(flow["path"]) - 1) * 10  # 3000 us per hop is the initial value
        self.dev_rtt[flow["flow_id"]] = 0.05 * self.estimated_rtt[flow["flow_id"]]  # 5% of the estimated RTT
        self.requests_in_flight[flow["flow_id"]] = []
        self.C[flow["flow_id"]] = self.default_C

    def delete_flow(self, flow_id):
        """
        Delete the flow from the congestion controller
        """
        del self.congestion_knobs[flow_id]
        del self.estimated_rtt[flow_id]
        del self.dev_rtt[flow_id]
        del self.requests_in_flight[flow_id]
        del self.other_ends[flow_id]
        del self.last_update[flow_id]
        del self.last_halved[flow_id]
        del self.ssthresh[flow_id]
        del self.is_slow_start[flow_id]

    def collect_timeouts(self, current_time):
        """
        Collect the timeouts for the requests in flight and update the congestion knobs for the involved flows
        """
        for flow_id in self.last_update:
            timed_out = False
            requests = self.requests_in_flight[flow_id]
            new_requests = []
            for req_id, time_sent, timeout in requests:
                if current_time - time_sent > timeout:
                    # the request has timed out
                    # we have to halve the congestion knob
                    timed_out = True
                else:
                    new_requests.append((req_id, time_sent, timeout))
            self.requests_in_flight[flow_id] = new_requests
            if timed_out:
                # sim_log.info(f"Flow {flow_id} timed out {len(requests) - len(new_requests)} times, halving congestion knob {self.congestion_knobs[flow_id]}")
                # self.halve_congestion_knob(flow_id, current_time)
                pass

    def handle_ack(self, flow_id, req_id, current_time, time_sent, mark_congested=False):
        """
        Handle the ACK for the given request

        Parameters
        ----------
        flow_id : int
            The flow id
        req_id : int
            The request id
        current_time : float
            The current time
        time_sent : float
            The time the request was sent
        mark_congested : bool, optional
            Whether the ack was marked as congested or not

        Returns
        -------
        int
            The number of new requests to generate for the flow
        """

        num_skipped = 0

        found = False

        # update the estimated RTT
        sample_rtt = current_time - time_sent
        self.estimated_rtt[flow_id] = 0.875 * self.estimated_rtt[flow_id] + 0.125 * sample_rtt
        self.dev_rtt[flow_id] = 0.75 * self.dev_rtt[flow_id] + 0.25 * abs(sample_rtt - self.estimated_rtt[flow_id])
        self.C[flow_id] = self.estimated_rtt[flow_id]*4000
        # print(f"Estimated RTT for flow {flow_id} is {self.estimated_rtt[flow_id]}, dev RTT is {self.dev_rtt[flow_id]}, C is {self.C}")

        for req, time_sent, timeout in self.requests_in_flight[flow_id]:
            if req < req_id:
                num_skipped += 1
                # if we receive an ACK for a request, we can remove all the requests with lower req_id
                # because they have been surely dropped due to congestion
            elif req == req_id:
                # we found the request for which we received the ACK
                # remove it from the list
                found = True

        if found:
            # remove the tuples from the list
            self.requests_in_flight[flow_id] = self.requests_in_flight[flow_id][num_skipped + 1:]

        else:
            self.requests_in_flight[flow_id] = self.requests_in_flight[flow_id][num_skipped:]
            """
            print(f"Flow {flow_id} received an ACK for a request that was not in flight, num_skipped={num_skipped}, "
                  f"req_id={req_id}, requests_in_flight={self.requests_in_flight[flow_id]}")
            """

        if mark_congested:
            sim_log.warning(f"Flow {flow_id} marked as congested")
            self.halve_congestion_knob(flow_id, current_time)

        # TODO: find another way to detect drops, e.g., timeouts
        """elif num_skipped > 0:
            sim_log.error(f"Flow {flow_id} marked as congested due to losses")
            self.halve_congestion_knob(flow_id, current_time)"""

        # return the number of new requests to generate. 0 with a rate-based controller
        return 0

    def get_inter_request_gap(self, flow_id):
        """
        Get the inter-request gap (us) for the given flow
        """
        return self.congestion_knobs[flow_id]

    def handle_new_request_in_flight(self, flow_id, req_id, current_time):
        """
        Handle a new request in flight for the given flow
        """
        timeout = max(self.estimated_rtt[flow_id] + 4 * self.dev_rtt[flow_id], 0.1)
        self.requests_in_flight[flow_id].append((req_id, current_time, timeout))

    def increase_all_knobs(self, current_time):
        for flow_id in self.last_update:
            self.increase_congestion_knob(flow_id, current_time)





def sanitize_flow_descriptors(flow_descriptors):
    """
    Sanitize the flow_descriptors parameter, transforming it into a dictionary indexed by flow_id
    """
    if isinstance(flow_descriptors, list):
        flow_descriptors_dict = {}
        for flow_descriptor in flow_descriptors:
            flow_descriptors_dict[flow_descriptor['flow_id']] = flow_descriptor
    elif isinstance(flow_descriptors, dict):
        flow_descriptors_dict = flow_descriptors
    else:
        raise ValueError("flow_descriptors must be a list or a dictionary")

    # check if the dictionary is well-formed
    for flow_id, flow_descriptor in flow_descriptors_dict.items():
        if 'flow_id' not in flow_descriptor:
            raise ValueError("flow_descriptor must contain a flow_id key")
        elif flow_descriptor['flow_id'] != flow_id:
            raise ValueError("flow_id in flow_descriptor does not match the key in the dictionary")

    return list(flow_descriptors_dict.values())



"""
Plot the rendezvous node index as a nice-looking histogram
"""

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

if __name__ == "__main__":
    fig, ax = plt.subplots()
    ax.set_title("Rendezvous Analysis")
    ax.set_xlabel("Rendezvous Node")
    ax.set_ylabel("Probability")
    df = pd.read_csv("./out2/rendezvous_node_vector.csv")
    # plot the histogram with a column for every value in df["sample"]
    # the height of the column is the number of occurrences of that value in df["sample"]

    # get the distinct values of the rendezvous node index
    distinct_values = df["sample"].unique()
    # sort the values
    distinct_values.sort()
    # compute the histogram
    hist = [df[df["sample"] == val].shape[0] for val in distinct_values]

    # normalize the histogram
    hist = np.array(hist) / np.sum(hist)

    # plot the histogram
    ax.bar(distinct_values, hist)

    plt.show()

    # save the figure as pdf
    fig.savefig("./out2/rendezvous_node_bars.pdf")



r"""
This module implements the :class:`~omnetpypy.front_end.port.Port` class.
Every port is an object that can receive and send messages. Ports are the interface of modules with the outside world.

Ports have the same semantic as in omnet++, except that
every port is an object itself and does not have to be input or output only.
"""


class Port:
    r"""
    This class represents a port in the simulation model.
    Ports are the interface of modules with the outside world.
    Every port is an object that can receive and send messages.

    Parameters
    ----------
    name : str
        The name of the port. This name should be unique within the parent module.
    parent : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
        The parent entity of the port, i.e., the entity to which the port belongs.

    Attributes
    ----------
    name : str
        The name of the port.
    connected_port : :class:`~omnetpypy.front_end.port.Port` or None
        The remote port to which this port is connected. If ``None``, the port is not connected.
        See also :meth:`~omnetpypy.front_end.port.Port.connect`.
    forwarded_input_port : :class:`~omnetpypy.front_end.port.Port` or None
        The remote port to which the input of this port is forwarded. If ``None``, the port does not forward input.
        See also :meth:`~omnetpypy.front_end.port.Port.forward_input`.
    forwarded_output_port : :class:`~omnetpypy.front_end.port.Port` or None
        The remote port to which the output of this port is forwarded. If ``None``, the port does not forward output.
        See also :meth:`~omnetpypy.front_end.port.Port.forward_output`.
    parent : :class:`~omnetpypy.front_end.sim_entity.SimulatedEntity`
        The parent entity of the port.
    subscribed_ports : list of :class:`~omnetpypy.front_end.port.Port`
        The list of ports that are subscribed to this port.
        Every subscribed port receives a copy of each message sent to this port.
    is_subscribed : bool
        A flag indicating whether this port is subscribed to another port.
        See also :meth:`~omnetpypy.front_end.port.Port.subscribe_to`.
    """

    def __init__(self, name, parent):
        self.name = name
        self.connected_port = None
        self.forwarded_input_port = None
        self.forwarded_output_port = None
        self.parent = parent
        self.subscribed_ports = []
        self.is_subscribed = False

    def connect(self, port):
        r"""
        Connect this port to another remote port.
        The output of this port will be fed as input to the other port, and vice versa.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The remote port to which this port will be connected.

        Raises
        ------
        ValueError
            If this port is already connected to another port,
            if the remote port is already connected to another port,
            if this port is already subscribed to another port,
            or if this port output is already forwarded to another port.
        """
        if self.is_subscribed:
            raise ValueError("Cannot connect a port that is already subscribed to another port")
        if port.connected_port is not None:
            raise ValueError("Cannot connect to a remote port that is already connected")
        if self.connected_port is not None:
            raise ValueError("Cannot connect a port that is already connected")
        if self.forwarded_output_port:
            raise ValueError("Cannot connect a port that is forwarding output")

        self.connected_port = port
        port.connected_port = self

    def subscribe_to(self, port):
        r"""
        Subscribe this port to another remote port.
        This port will receive a copy of each message sent to the remote port.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The remote port to which this port will be subscribed.

        Raises
        ------
        ValueError
            If this port is already connected to another port,
            if the remote port is already connected to another port,
            or if the remote port is forwarding output
        """

        if self.connected_port is not None:
            raise ValueError("Cannot subscribe a port that is already connected")
        if port.forwarded_output_port:
            raise ValueError("Cannot subscribe to a port that is forwarding output")
        if port.connected_port is not None:
            raise ValueError("Cannot subscribe to a port that is already connected")

        port.subscribed_ports.append(self)
        self.is_subscribed = True

    def disconnect(self, port):
        r"""
        Disconnect this port from another remote port.
        If the two ports are not connected to any other port, this does nothing.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The remote port from which this port will be disconnected.

        Raises
        ------
        ValueError
            If this or the other port are connected to a third port and not mutually connected.
        """

        # check whether the two ports were actually connected
        if self.connected_port != port or port.connected_port != self:
            raise ValueError("The two ports are not connected")

        self.connected_port = None
        port.connected_port = None

    def forward_input(self, port):
        r"""
        Forward the input of this port to another remote port.
        The input of this port will be sent as input to the other port.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The remote port to which the input of this port will be forwarded.

        Raises
        ------
        ValueError
            If this port is already connected to another port,
            or if the remote port is already connected to another port,
        """

        if self.connected_port is not None:
            raise ValueError("Cannot forward input from a port that is already connected")
        if port.connected_port is not None:
            raise ValueError("Cannot forward input to a port that is already connected")

        self.forwarded_input_port = port

    def forward_output(self, port):
        r"""
        Forward the output of this port to another remote port.
        The output of this port will be sent as output to the other port.

        Parameters
        ----------
        port : :class:`~omnetpypy.front_end.port.Port`
            The remote port to which the output of this port will be forwarded.

        Raises
        ------
        ValueError
            If this port is already connected to another port,
            or if the remote port is already connected to another port,
        """

        if self.connected_port is not None:
            raise ValueError("Cannot forward input from a port that is already connected")
        if port.connected_port is not None:
            raise ValueError("Cannot forward input to a port that is already connected")

        self.forwarded_output_port = port

    def tx_output(self, message):
        r"""
        Send a message as output of this port.

        Parameters
        ----------
        message : :class:`~omnetpypy.front_end.message.Message`
            The message to be sent.
        """

        if self.forwarded_output_port:
            self.forwarded_output_port.tx_output(message)
        elif self.connected_port:
            self.connected_port.tx_input(message)
        elif len(self.subscribed_ports) > 0:
            for port in self.subscribed_ports:
                message_cpy = message.__copy__()
                port.tx_input(message_cpy)
        else:
            # do nothing, the mesaage is lost
            pass

    def tx_input(self, message):
        r"""
        Receive a message as input of this port.

        Parameters
        ----------
        message : :class:`~omnetpypy:front_end.message.Message`
            The message to be received.
        """

        if self.forwarded_input_port:
            self.forwarded_input_port.tx_input(message)
        elif self.parent.is_listening:
            self.parent.sim_context.connector.schedule_port_input(self, message)
        else:
            # do nothing, the mesaage is lost
            pass



"""
Print a plot of the queuing_time_vector.csv file as a sliding window average with window size 1000 time units.
"""

import pandas as pd
import matplotlib.pyplot as plt

# Load the data
df = pd.read_csv("./out/queuing_time_vector.csv")

# Compute the sliding window average
window_size_time_units = 100

# create a new column with the sliding window index
df["sw_index"] = df["timestamp"] // window_size_time_units

# compute the sliding window average
sw_avg = df.groupby("sw_index").mean()

# divide the timestamp and the sample by 1e3 to have ms
sw_avg["timestamp"] /= 1e3
sw_avg["sample"] /= 1e3

# plot the sliding window average as a function of timestamp
plt.plot(sw_avg["timestamp"], sw_avg["sample"])
plt.xlabel("Time (ms)")
plt.ylabel("Queuing time (ms)")
plt.title("Moving average of queuing time")

plt.show()



r"""
An omnet++-like, user-friendly simulation package fully implemented on Python,
with the possibility to support different simulation backends such as SimPy or PyDynaa while keeping the same front-end.

The package is designed to be as similar in concept and usage to the omnet++ simulation framework, but with the
advantages of being written in Python, which makes it more flexible and easy to use.
"""

from omnetpypy.front_end import SimpleModule, Message, CompoundModule, Channel
from omnetpypy.simulation import Experiment

__all__ = ["SimpleModule", "Message", "CompoundModule", "Channel", "Experiment"]
